﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.IO;
using Newtonsoft.Json;
using QRCoder;
using MVB_RestBookService.CodeLists;
using System.Globalization;



namespace MVB_RestBookService
{
    public partial class MainFrame : Form
    {
        //Enums
        private enum DATABASES { VLB, SBZ, OESB };

        //Access
        private string RBSPassword = "";
		private string account = "";
        private string password = "";
        private string statefullAccessToken = "";

        //Codelists
        private CodeList_MediaData codeList_MediaData;
        private CodeList_005_identifier codeList_005_identifier;
        private CodeList_007_productform codeList_007_productform;
        private CodeList_017_contributorRole codeList_017_contributorRole;
        private CodeList_027_subjectIdentifier codeList_027_subjectIdentifier;
        private CodeList_037_mediaFileFormat codeList_037_mediaFileFormat;
        private CodeList_038_mediaFileType codeList_038_mediaFileType;
        private CodeList_064_publishingStatus codeList_064_publishingStatus;
        private CodeList_065_availability codeList_065_availability;
        private CodeList_074_language codeList_074_language;

        //Search Parameters
        private string searchParam_quickSearch;
		private string searchParam_searchByISBN;
		private string searchParam_searchByTitle;
		private string searchParam_searchByAuthor;
		private string searchParam_searchByPublisher;
		private string searchParam_searchBySubject;
        private string searchParam_searchBySet;
        private string searchParam_searchByCollection;
        private string searchParam_searchByProductForm;
        private string searchParam_searchByLanguage;
        private string searchParam_searchByPriceFrom;
        private string searchParam_searchByPriceTo;
        private string searchParam_searchByPublicationDateFrom;
        private string searchParam_searchByPublicationDateTo;
        private DATABASES searchParam_database;

        private string sortField = "title"; //isbn, title, author, publisher, price
		private string sortDirection = "asc"; //asc, desc
        private int page = 1;
        private int lastPage = 1;
        private int amountOfTitlesShown = 50;

        private bool atLeastOneSearchPerformed = false;

        //Detail View Parameters
        private Font listItemDefaultFontRegular = new Font("Microsoft Sans Serif", 8.25f, FontStyle.Regular, GraphicsUnit.Point, ((byte)(0)));
        private Font listItemDefaultFontBold = new Font("Microsoft Sans Serif", 8.25f, FontStyle.Bold, GraphicsUnit.Point, ((byte)(0)));
        private string currentPublisherMVBID = "";

        //QRCode Parameters
        private string currentlyShownISBN = "";
        private string currentlyShownID = "";

        //Cover
        Image cover;

        //Debug
        bool debugMode = false;



        public MainFrame()
        {
            InitializeComponent();

            //Debug
            if (!debugMode)
            {
                tabControl.TabPages.RemoveAt(tabControl.TabPages.Count - 1);
            }

            if (debugMode)
            {
                waterMarkTextBox_search_QuickSearch.Text = "9783868892666";
            }

            //Init
            setPageButtonStatus();
            comboBox_search_sortingField.SelectedIndex = 0;
            comboBox_search_sortingDirection.SelectedIndex = 0;
            comboBox_search_itemsPerPage.SelectedIndex = 2;
            comboBox_search_database.SelectedIndex = 0;
            tabControl_selectedIndexChange(null, null);

            //Init Tab Pages (a small Part of this Init can't be done by the UI-Designer)
            initTitleDetailViewPage();
            initPublisherDetailPage();

            //Listeners
            FormClosing += MainFrame_FormClosing;

            //Init CodeLists
            codeList_MediaData = new CodeList_MediaData();
            codeList_005_identifier = new CodeList_005_identifier();
            codeList_007_productform = new CodeList_007_productform();
            codeList_017_contributorRole = new CodeList_017_contributorRole();
            codeList_027_subjectIdentifier = new CodeList_027_subjectIdentifier();
            codeList_037_mediaFileFormat = new CodeList_037_mediaFileFormat();
            codeList_038_mediaFileType = new CodeList_038_mediaFileType();
            codeList_064_publishingStatus = new CodeList_064_publishingStatus();
            codeList_065_availability = new CodeList_065_availability();
            codeList_074_language = new CodeList_074_language();

            //Ask for Password
            if (debugMode)
            {
                RBSPassword = "piep";
            }
            else
            {
                LoginFrame loginFrame = new LoginFrame();
                loginFrame.ShowDialog();
                RBSPassword = loginFrame.password;
            }

            //Read Login Data from File
            try
            {
                if (File.Exists("loginData.conf"))
                {
                    //Open Stream
                    FileStream fileStream = new FileStream("loginData.conf", FileMode.Open);
                    StreamReader streamReader = new StreamReader(fileStream);

                    //Variables / Read Data
                    string encryptedAccount = streamReader.ReadLine();
                    string encryptedPasswort = streamReader.ReadLine();
                    string encryptedMetadataToken = streamReader.ReadLine();
                    string encryptedCoverToken = streamReader.ReadLine();
                    string encryptedMediaToken = streamReader.ReadLine();

                    //Close Stream
                    streamReader.Close();
                    fileStream.Close();

                    //Decrypt Data
                    account = encryptedAccount.Equals("") ? "" : Encryption.DecryptStringAES(encryptedAccount, RBSPassword);
                    password = encryptedPasswort.Equals("") ? "" : Encryption.DecryptStringAES(encryptedPasswort, RBSPassword);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Ein Fehler ist aufgetreten:\n\nDas eingegebene Passwort ist nicht korrekt.", "Fehler");
                return;
            }

            //Missing Account or Password
            if(account.Equals("") || password.Equals(""))
            {
                MessageBox.Show("Sie haben bislang keine Zugangsdaten eingegeben.\nOhne Zugangsdaten können keine Informationen über das VLB abgefragt werden.\n\n", "Info");
                return;
            }

            MessageBox.Show("Das Programm loggt sich im VLB ein. Dies kann bis zu einer Minute in Anspruch nehmen.", "Info");

            //Login with Login Data
            RestClient client = new RestClient();
            client.EndPoint = @"http://api.vlb.de/api/v1/login";
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.POST;
            client.PostData = "{ \"username\": \"" + account.Trim() + "\", \"password\": \"" + password.Trim() + "\" }";

            //Send Request
            try
            {
                client.MakeRequest("", ref statefullAccessToken);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                return;
            }

            MessageBox.Show("Sie wurden erfolgreich eingeloggt.", "Info");
        }



        /// <summary>
        /// Init Title Detail View.
        /// </summary>
        private void initTitleDetailViewPage()
        {
            label_detail_subTitle.Text = "-";
            label_detail_authors.Text = "-";
            label_detail_priceEURD.Text = "-";
            label_detail_priceEURA.Text = "-";
            label_detail_priceCHF.Text = "-";
            label_detail_ISBN.Text = "-";
            label_detail_publisher.Text = "-";
            label_detail_produktform.Text = "-";
            label_detail_height.Text = "-";
            label_detail_width.Text = "-";
            label_detail_weight.Text = "-";
            label_detail_pages.Text = "-";
        }



        /// <summary>
        /// Init Publisher Detail View (a small Part of this Init can't be done by the UI-Designer).
        /// </summary>
        private void initPublisherDetailPage()
        {
            //Set Links for LinkLabels
            linkLabel_publisher_email.Links[0].Name = "Email";
            linkLabel_publisher_website.Links[0].Name = "URL";
            linkLabel_publisher_isbnPrefixes1.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes2.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes3.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes4.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes5.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes6.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes7.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes8.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes9.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes10.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes11.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes12.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes13.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes14.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes15.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes16.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes17.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes18.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes19.Links[0].Name = "IS*";
            linkLabel_publisher_isbnPrefixes20.Links[0].Name = "IS*";

            //Clear all Data
            richTextBox_publisher_mvbid.Text = "-";
            richTextBox_publisher_street.Text = "-";
            richTextBox_publisher_cityStreet.Text = "-";
            richTextBox_publisher_zipStreet.Text = "-";
            richTextBox_publisher_postbox.Text = "-";
            richTextBox_publisher_zipPostbox.Text = "-";
            richTextBox_publisher_phone.Text = "-";
            richTextBox_publisher_fax.Text = "-";
            linkLabel_publisher_email.Text = "";
            linkLabel_publisher_website.Text = "";
            richTextBox_publisher_deliveryBS.Text = "-";
            linkLabel_publisher_isbnPrefixes1.Text = "";
            linkLabel_publisher_isbnPrefixes2.Text = "";
            linkLabel_publisher_isbnPrefixes3.Text = "";
            linkLabel_publisher_isbnPrefixes4.Text = "";
            linkLabel_publisher_isbnPrefixes5.Text = "";
            linkLabel_publisher_isbnPrefixes6.Text = "";
            linkLabel_publisher_isbnPrefixes7.Text = "";
            linkLabel_publisher_isbnPrefixes8.Text = "";
            linkLabel_publisher_isbnPrefixes9.Text = "";
            linkLabel_publisher_isbnPrefixes10.Text = "";
            linkLabel_publisher_isbnPrefixes11.Text = "";
            linkLabel_publisher_isbnPrefixes12.Text = "";
            linkLabel_publisher_isbnPrefixes13.Text = "";
            linkLabel_publisher_isbnPrefixes14.Text = "";
            linkLabel_publisher_isbnPrefixes15.Text = "";
            linkLabel_publisher_isbnPrefixes16.Text = "";
            linkLabel_publisher_isbnPrefixes17.Text = "";
            linkLabel_publisher_isbnPrefixes18.Text = "";
            linkLabel_publisher_isbnPrefixes19.Text = "";
            linkLabel_publisher_isbnPrefixes20.Text = "";
        }



        /// <summary>
        /// Starts a new Search with the set Parameters.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_startSearch_Click(object sender, EventArgs e)
        {
            //Abort, if Search is empty
            if (waterMarkTextBox_search_QuickSearch.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByISBN.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByTitle.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByAuthor.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByPublisher.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchBySubject.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchBySet.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByCollection.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByProductForm.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByLanguage.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByPriceFrom.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByPriceTo.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByPublicationDateFrom.Text.Trim().Equals("") &&
                waterMarkTextBox_search_searchByPublicationDateTo.Text.Trim().Equals(""))
            {
                dataGridView_search_results.Rows.Clear();
                page = 1;
                lastPage = 1;
                setPageButtonStatus();
                toolStripStatusLabel.Text = "Kein Suchbegriff eingegeben!";
                return;
            }

            //Save Search Parameters in Buffer
            searchParam_quickSearch = waterMarkTextBox_search_QuickSearch.Text.Trim();
            searchParam_searchByISBN = waterMarkTextBox_search_searchByISBN.Text.Trim();
            searchParam_searchByTitle = waterMarkTextBox_search_searchByTitle.Text.Trim();
            searchParam_searchByAuthor = waterMarkTextBox_search_searchByAuthor.Text.Trim();
            searchParam_searchByPublisher = waterMarkTextBox_search_searchByPublisher.Text.Trim();
            searchParam_searchBySubject = waterMarkTextBox_search_searchBySubject.Text.Trim();
            searchParam_searchBySet = waterMarkTextBox_search_searchBySet.Text.Trim();
            searchParam_searchByCollection = waterMarkTextBox_search_searchByCollection.Text.Trim();
            searchParam_searchByProductForm = waterMarkTextBox_search_searchByProductForm.Text.Trim();
            searchParam_searchByLanguage = waterMarkTextBox_search_searchByLanguage.Text.Trim();
            searchParam_searchByPriceFrom = waterMarkTextBox_search_searchByPriceFrom.Text.Trim().Replace(",", ".");
            searchParam_searchByPriceTo = waterMarkTextBox_search_searchByPriceTo.Text.Trim().Replace(",", ".");
            searchParam_searchByPublicationDateFrom = waterMarkTextBox_search_searchByPublicationDateFrom.Text.Trim();
            searchParam_searchByPublicationDateTo = waterMarkTextBox_search_searchByPublicationDateTo.Text.Trim();


            //Check, if Search Parameters are valid
            float notUsedFloat;
            int notUsedInt;

            if (searchParam_searchByPriceFrom.Equals("") ? false : !float.TryParse(searchParam_searchByPriceFrom, out notUsedFloat) || searchParam_searchByPriceTo.Equals("") ? false : !float.TryParse(searchParam_searchByPriceTo, out notUsedFloat))
            {
                MessageBox.Show("Achtung:\n\nEs wurde möglicherweise ein ungültiger Preis angegeben.\nBitte verwenden Sie ausschließlich Ziffern mit durch Komma oder Punkt\ngetrennte Cent-Beträge.", "Info");
            }

            if (searchParam_searchByPublicationDateFrom.Equals("") ? false : !int.TryParse(searchParam_searchByPublicationDateFrom, out notUsedInt) || searchParam_searchByPublicationDateTo.Equals("") ? false : !int.TryParse(searchParam_searchByPublicationDateTo, out notUsedInt))
            {
                MessageBox.Show("Achtung:\n\nEs wurde möglicherweise ein ungültiges Jahr angegeben.\nBitte verwenden Sie ausschließlich Ziffern.\nGeben Sie keine Tage oder Monate an.", "Info");
            }


            //Perform Search
            page = 1;

            performSearch();

            atLeastOneSearchPerformed = true;
        }



        /// <summary>
        /// Fired, if the user double-clicks at a title in the search_dataGridView.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void search_dataGridView_onDoubleClick(object sender, DataGridViewCellMouseEventArgs e)
        {
            //Header was doubleClicked, abort
            if (e.RowIndex < 0)
            {
                return;
            }

            //Variables
            RestClient client;
            HttpWebResponse json;
            string isbn;
            string responseAsString = null;

            //No ISBN available
            if(dataGridView_search_results[0, e.RowIndex].Value == null || dataGridView_search_results[0, e.RowIndex].Value.Equals(""))
            {
                MessageBox.Show("Der gesuchte Titel besitzt keine ISBN, weshalb keine Detailansicht möglich ist.", "Fehler");
                return;
            }

            //Get Title clicked and send REST Request
            isbn = dataGridView_search_results[0, e.RowIndex].Value.ToString();

            client = new RestClient();
            client.EndPoint = @"http://api.vlb.de/api/v1/product/" + isbn.Replace("-", "") + @"/isbn13";
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.GET;

            //Send Request
            try
            {
                json = client.MakeRequest("", ref responseAsString);
            }
            catch (Exception ex)
            {
                toolStripStatusLabel.Text = String.Format("Fehler: {0}", ex.Message);
                return;
            }

            //Init Json.net Deserializer
            JsonSerializerSettings jsonSerializerSettings = new JsonSerializerSettings();
            jsonSerializerSettings.MissingMemberHandling = MissingMemberHandling.Ignore;
            jsonSerializerSettings.NullValueHandling = NullValueHandling.Ignore;

            //Debug Output
            if (!checkBox_debug_appendText.Checked)
            {
                richTextBox_debug_rawData.Clear();
            }
            
            richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
            richTextBox_debug_rawData.AppendText("Response:\n" + JsonFormatter.FormatJson(responseAsString) + "\n\n");

            //Create a Container, containing all Json Fields and deserialize Json Data
            JsonDataContainerDetail jsonDataContainerDetailRaw = JsonConvert.DeserializeObject<JsonDataContainerDetail>(responseAsString, jsonSerializerSettings);
            JsonDataContainerDetail jsonDataContainerDetail = Transformer.transform(jsonDataContainerDetailRaw);

            //Get Media Data
            client = new RestClient();
            client.EndPoint = @"http://api.vlb.de/api/v1/asset/mmo/" + jsonDataContainerDetail.id;
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.GET;

            //Send Request
            JsonDataContainerMedia[] jsonDataContainerMedia = null;

            try
            {
                json = client.MakeRequest("", ref responseAsString);

                richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
                richTextBox_debug_rawData.AppendText("Response:\n" + JsonFormatter.FormatJson(responseAsString) + "\n\n");

                jsonDataContainerMedia = JsonConvert.DeserializeObject<JsonDataContainerMedia[]>(responseAsString, jsonSerializerSettings);
            }
            catch (Exception ex)
            {
                //toolStripStatusLabel.Text = String.Format("Fehler: {0}", ex.Message);
            }

            //Set Title Detail View

            //Title
            #region title
            {
                label_detail_title.Text = jsonDataContainerDetail.titles[0].title;
            }
            #endregion

            //Subtitle
            #region subtitle
            {
                label_detail_subTitle.Text = jsonDataContainerDetail.titles[0].subtitle;
            }
            #endregion

            //Authors
            #region authors
            {
                //Variables
                List<string> authors = new List<string>();

                //Get Contributor Names
                for (int i = 0; i < jsonDataContainerDetail.contributors.Length; i++)
                {
                    //Get Contributor Role
                    string contributorRole = codeList_017_contributorRole.getCodeListValue(jsonDataContainerDetail.contributors[i].type);

                    authors.Add(String.Format("{0} {1} {2}", contributorRole, jsonDataContainerDetail.contributors[i].firstName, jsonDataContainerDetail.contributors[i].lastName));
                }

                label_detail_authors.Text = string.Join(", ", authors.ToArray<string>());
            }
            #endregion

            //Prices
            #region prices
            {
                //EUR D
                {
                    label_detail_priceEURD.Text = "- EUR (DE)";

                    //Search Prices Composite for appropriate Country
                    for (int i = 0; i < jsonDataContainerDetail.prices.Length; i++)
                    {
                        //Set Price
                        if (jsonDataContainerDetail.prices[i].country.Equals("DE", StringComparison.CurrentCultureIgnoreCase))
                        {
                            //Variables
                            string provisional = "";
                            string fixedRetailPrice = "";
                            string taxRate = "";

                            //Provisional
                            if (!jsonDataContainerDetail.prices[i].provisional.Equals(""))
                            {
                                provisional = string.Format("{0} ", jsonDataContainerDetail.prices[i].provisional);
                            }

                            //Taxation Description
                            if (!jsonDataContainerDetail.prices[i].taxRate.Equals(""))
                            {
                                taxRate = string.Format(", {0}", jsonDataContainerDetail.prices[i].taxRate);
                            }

                            //Fixed Price
                            if (!jsonDataContainerDetail.prices[i].fixedRetailPrice.Equals(""))
                            {
                                fixedRetailPrice = string.Format(", {0}", jsonDataContainerDetail.prices[i].fixedRetailPrice);
                            }

                            label_detail_priceEURD.Text = String.Format("{1}{0} EUR (DE){2}{3}", jsonDataContainerDetail.prices[i].value, provisional, taxRate, fixedRetailPrice);
                        }
                    }
                }

                //EUR A
                {
                    label_detail_priceEURA.Text = "- EUR (AT)";

                    //Search Prices Composite for appropriate Country
                    for (int i = 0; i < jsonDataContainerDetail.prices.Length; i++)
                    {
                        //Set Price
                        if (jsonDataContainerDetail.prices[i].country.Equals("AT", StringComparison.CurrentCultureIgnoreCase))
                        {
                            //Variables
                            string provisional = "";
                            string fixedRetailPrice = "";
                            string taxRate = "";

                            //Provisional
                            if (!jsonDataContainerDetail.prices[i].provisional.Equals(""))
                            {
                                provisional = string.Format("{0} ", jsonDataContainerDetail.prices[i].provisional);
                            }

                            //Taxation Description
                            if (!jsonDataContainerDetail.prices[i].taxRate.Equals(""))
                            {
                                taxRate = string.Format(", {0}", jsonDataContainerDetail.prices[i].taxRate);
                            }

                            //Fixed Price
                            if (!jsonDataContainerDetail.prices[i].fixedRetailPrice.Equals(""))
                            {
                                fixedRetailPrice = string.Format(", {0}", jsonDataContainerDetail.prices[i].fixedRetailPrice);
                            }

                            label_detail_priceEURA.Text = String.Format("{1}{0} EUR (AT){2}{3}", jsonDataContainerDetail.prices[i].value, provisional, taxRate, fixedRetailPrice);
                        }
                    }

                }

                //CHF
                {
                    label_detail_priceCHF.Text = "- CHF";

                    //Search Prices Composite for appropriate Country
                    for (int i = 0; i < jsonDataContainerDetail.prices.Length; i++)
                    {
                        //Set Price
                        if (jsonDataContainerDetail.prices[i].country.Equals("CH", StringComparison.CurrentCultureIgnoreCase))
                        {
                            //Variables
                            string provisional = "";
                            string fixedRetailPrice = "";
                            string taxRate = "";

                            //Provisional
                            if (!jsonDataContainerDetail.prices[i].provisional.Equals(""))
                            {
                                provisional = string.Format("{0} ", jsonDataContainerDetail.prices[i].provisional);
                            }

                            //Taxation Description
                            if (!jsonDataContainerDetail.prices[i].taxRate.Equals(""))
                            {
                                taxRate = string.Format(", {0}", jsonDataContainerDetail.prices[i].taxRate);
                            }

                            //Fixed Price
                            if (!jsonDataContainerDetail.prices[i].fixedRetailPrice.Equals(""))
                            {
                                fixedRetailPrice = string.Format(", {0}", jsonDataContainerDetail.prices[i].fixedRetailPrice);
                            }

                            label_detail_priceCHF.Text = String.Format("{1}{0} CHF{2}{3}", jsonDataContainerDetail.prices[i].value, provisional, taxRate, fixedRetailPrice);
                        }
                    }

                }
            }
            #endregion

            //ISBN
            #region isbn
            {
                //Variables
                string isbn13 = "-";

                //Get ISBN
                for (int i = 0; i < jsonDataContainerDetail.identifiers.Length; i++)
                {
                    if (jsonDataContainerDetail.identifiers[i].type.Equals("15", StringComparison.CurrentCultureIgnoreCase))
                    {
                        isbn13 = jsonDataContainerDetail.identifiers[i].value;
                    }
                }

                label_detail_ISBN.Text = isbn13;
            }
            #endregion

            //Publisher
            #region publisher
            {
                //Variables
                string publisher = "-";

                //Get ISBN
                for (int i = 0; i < jsonDataContainerDetail.publishers.Length; i++)
                {
                    if (jsonDataContainerDetail.publishers[i].type.Equals("01", StringComparison.CurrentCultureIgnoreCase))
                    {
                        publisher = jsonDataContainerDetail.publishers[i].name;
                    }
                }

                label_detail_publisher.Text = publisher;
            }
            #endregion

            //Productform
            #region productform
            {
                //Variables
                string type = "-";
                string binding = "";

                //Get Type
                type = codeList_007_productform.getCodeListValue(jsonDataContainerDetail.form.type);

                //Get Binding
                binding = String.Format(" ({0})", jsonDataContainerDetail.form.binding);

                label_detail_produktform.Text = String.Format("{0}{1}", type, binding);
            }
            #endregion

            //Height
            #region height
            {
                label_detail_height.Text = "-";

                if (!jsonDataContainerDetail.form.height.Equals(""))
                {
                    label_detail_height.Text = String.Format("{0} mm", jsonDataContainerDetail.form.height);
                }
            }
            #endregion

            //Width
            #region width
            {
                label_detail_width.Text = "-";

                if (!jsonDataContainerDetail.form.width.Equals(""))
                {
                    label_detail_width.Text = String.Format("{0} mm", jsonDataContainerDetail.form.width);
                }
            }
            #endregion

            //Weight
            #region weight
            {
                label_detail_weight.Text = "-";

                if (!jsonDataContainerDetail.form.weight.Equals(""))
                {
                    label_detail_weight.Text = String.Format("{0} g", jsonDataContainerDetail.form.weight);
                }
            }
            #endregion

            //Pages
            #region pages
            {
                label_detail_pages.Text = "-";

                if (!jsonDataContainerDetail.extent.pages.Equals(""))
                {
                    label_detail_pages.Text = String.Format("ca. {0} Seiten", jsonDataContainerDetail.extent.pages);
                }
            }
            #endregion

            //Set Detail Tabs Data

            //Authors
            #region authors
            {
                //Variables
                List<Contributors> contributors = new List<Contributors>();
                int countAuthors = 0;
                int countTranslators = 0;
                int countIllustrators = 0;
                int countOtherContributors = 0;

                //Real Authors
                for (int i = 0; i < jsonDataContainerDetail.contributors.Length; i++)
                {
                    if (jsonDataContainerDetail.contributors[i].type.Equals("A01", StringComparison.CurrentCultureIgnoreCase))
                    {
                        contributors.Add(jsonDataContainerDetail.contributors[i]);
                        countAuthors++;
                    }
                }

                //Translators
                for (int i = 0; i < jsonDataContainerDetail.contributors.Length; i++)
                {
                    if (jsonDataContainerDetail.contributors[i].type.Equals("B06", StringComparison.CurrentCultureIgnoreCase))
                    {
                        contributors.Add(jsonDataContainerDetail.contributors[i]);
                        countTranslators++;
                    }
                }

                //Illustrators
                for (int i = 0; i < jsonDataContainerDetail.contributors.Length; i++)
                {
                    if (jsonDataContainerDetail.contributors[i].type.Equals("A12", StringComparison.CurrentCultureIgnoreCase))
                    {
                        contributors.Add(jsonDataContainerDetail.contributors[i]);
                        countIllustrators++;
                    }
                }

                //Other Contributors
                for (int i = 0; i < jsonDataContainerDetail.contributors.Length; i++)
                {
                    if (!jsonDataContainerDetail.contributors[i].type.Equals("A01", StringComparison.CurrentCultureIgnoreCase) && !jsonDataContainerDetail.contributors[i].type.Equals("B06", StringComparison.CurrentCultureIgnoreCase) && !jsonDataContainerDetail.contributors[i].type.Equals("A12", StringComparison.CurrentCultureIgnoreCase))
                    {
                        contributors.Add(jsonDataContainerDetail.contributors[i]);
                        countOtherContributors++;
                    }
                }

                //Init Table Data
                string[,] data = new string[6, contributors.Count + 1];
                Font[,] font = new Font[6, contributors.Count + 1];
                string[,] hyperlinkValue = new string[6, contributors.Count + 1];
                int[] columnWidth = new int[6];
                bool[] columnWordWrap = new bool[6];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 100;
                columnWidth[1] = 200;
                columnWidth[2] = 100;
                columnWidth[3] = 100;
                columnWidth[4] = 100;
                columnWidth[5] = 150;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;
                columnWordWrap[2] = false;
                columnWordWrap[3] = false;
                columnWordWrap[4] = false;
                columnWordWrap[5] = true;

                //Add Header
                data[0, 0] = "";
                data[1, 0] = "Name";
                data[2, 0] = "Sprache";
                data[3, 0] = "Geboren am";
                data[4, 0] = "Gestorben am";
                data[5, 0] = "Website(s)";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                font[2, 0] = listItemDefaultFontBold;
                font[3, 0] = listItemDefaultFontBold;
                font[4, 0] = listItemDefaultFontBold;
                font[5, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";
                hyperlinkValue[2, 0] = "";
                hyperlinkValue[3, 0] = "";
                hyperlinkValue[4, 0] = "";
                hyperlinkValue[5, 0] = "";

                //Add Data to Table
                for (int i = 0; i < contributors.Count; i++)
                {
                    data[0, i + 1] = "";
                    font[0, i + 1] = listItemDefaultFontBold;
                    hyperlinkValue[0, i + 1] = "";

                    //Row Header
                    if (countAuthors != 0 && i == 0)
                    {
                        data[0, i + 1] = "Autor(en)";
                    }

                    if (countTranslators != 0 && i == countAuthors)
                    {
                        data[0, i + 1] = "Übersetzer";
                    }

                    if (countIllustrators != 0 && i == countAuthors + countTranslators)
                    {
                        data[0, i + 1] = "Illustrator(en)";
                    }

                    if (countOtherContributors != 0 && i == countAuthors + countTranslators + countIllustrators)
                    {
                        data[0, i + 1] = "Andere";
                    }

                    //Data
                    data[1, i + 1] = contributors[i].fullName;
                    data[2, i + 1] = contributors[i].language;
                    data[3, i + 1] = contributors[i].dateOfBirth;
                    data[4, i + 1] = contributors[i].dateOfDeath;
                    data[5, i + 1] = "";

                    for (int j = 0; j < contributors[i].webSites.Length; j++)
                    {
                        data[5, i + 1] += contributors[i].webSites[j].url;

                        if (j != contributors[i].webSites.Length - 1)
                        {
                            data[5, i + 1] += "\n";
                        }
                    }

                    font[1, i + 1] = listItemDefaultFontRegular;
                    font[2, i + 1] = listItemDefaultFontRegular;
                    font[3, i + 1] = listItemDefaultFontRegular;
                    font[4, i + 1] = listItemDefaultFontRegular;
                    font[5, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[1, i + 1] = "AU";
                    hyperlinkValue[2, i + 1] = "";
                    hyperlinkValue[3, i + 1] = "";
                    hyperlinkValue[4, i + 1] = "";
                    hyperlinkValue[5, i + 1] = "URL";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[0].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[0].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Publishers
            #region publishers
            {
                //Variables
                List<Publishers> publishers = new List<Publishers>();
                int countPublishers = 0;
                int countCoPublishers = 0;
                int countOtherPublishers = 0;

                //Publishers
                for (int i = 0; i < jsonDataContainerDetail.publishers.Length; i++)
                {
                    if (jsonDataContainerDetail.publishers[i].type.Equals("01", StringComparison.CurrentCultureIgnoreCase))
                    {
                        publishers.Add(jsonDataContainerDetail.publishers[i]);
                        countPublishers++;
                    }
                }

                //Co-Publishers
                for (int i = 0; i < jsonDataContainerDetail.publishers.Length; i++)
                {
                    if (jsonDataContainerDetail.publishers[i].type.Equals("02", StringComparison.CurrentCultureIgnoreCase))
                    {
                        publishers.Add(jsonDataContainerDetail.publishers[i]);
                        countCoPublishers++;
                    }
                }

                //Other Publishers
                for (int i = 0; i < jsonDataContainerDetail.publishers.Length; i++)
                {
                    if (!jsonDataContainerDetail.publishers[i].type.Equals("01", StringComparison.CurrentCultureIgnoreCase) && !jsonDataContainerDetail.publishers[i].type.Equals("02", StringComparison.CurrentCultureIgnoreCase))
                    {
                        publishers.Add(jsonDataContainerDetail.publishers[i]);
                        countOtherPublishers++;
                    }
                }

                //Init Table Data
                string[,] data = new string[4, publishers.Count + 1];
                Font[,] font = new Font[4, publishers.Count + 1];
                string[,] hyperlinkValue = new string[4, publishers.Count + 1];
                int[] columnWidth = new int[4];
                bool[] columnWordWrap = new bool[4];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 100;
                columnWidth[1] = 300;
                columnWidth[2] = 100;
                columnWidth[3] = 150;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;
                columnWordWrap[2] = false;
                columnWordWrap[3] = true;

                //Add Header
                data[0, 0] = "";
                data[1, 0] = "Name";
                data[2, 0] = "MVB-Kennnr.";
                data[3, 0] = "Website(s)";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                font[2, 0] = listItemDefaultFontBold;
                font[3, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";
                hyperlinkValue[2, 0] = "";
                hyperlinkValue[3, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < publishers.Count; i++)
                {
                    data[0, i + 1] = "";
                    font[0, i + 1] = listItemDefaultFontBold;
                    hyperlinkValue[0, i + 1] = "";

                    //Row Header
                    if (countPublishers != 0 && i == 0)
                    {
                        data[0, i + 1] = "Verlag(e)";
                        currentPublisherMVBID = publishers[i].publisherId;
                    }

                    if (countCoPublishers != 0 && i == countPublishers)
                    {
                        data[0, i + 1] = "Co-Verlag(e)";
                    }

                    if (countOtherPublishers != 0 && i == countPublishers + countCoPublishers)
                    {
                        data[0, i + 1] = "Andere";
                    }

                    //Data
                    data[1, i + 1] = publishers[i].name;
                    data[2, i + 1] = publishers[i].publisherId;
                    data[3, i + 1] = "";

                    for (int j = 0; j < publishers[i].webSites.Length; j++)
                    {
                        data[3, i + 1] += publishers[i].webSites[j].url;

                        if (j != publishers[i].webSites.Length - 1)
                        {
                            data[3, i + 1] += "\n";
                        }
                    }

                    font[1, i + 1] = listItemDefaultFontRegular;
                    font[2, i + 1] = listItemDefaultFontRegular;
                    font[3, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[1, i + 1] = "";
                    hyperlinkValue[2, i + 1] = "VL";
                    hyperlinkValue[3, i + 1] = "URL";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[1].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[1].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Texts
            #region texts
            {
                //Variables
                List<Texts> texts = new List<Texts>();
                int countMainDescription = 0;
                int countShortDescription = 0;
                int countLongDescription = 0;
                int countTableOfContents = 0;
                int countOtherTexts = 0;

                //Main Description
                for (int i = 0; i < jsonDataContainerDetail.texts.Length; i++)
                {
                    if (jsonDataContainerDetail.texts[i].type.Equals("01", StringComparison.CurrentCultureIgnoreCase))
                    {
                        texts.Add(jsonDataContainerDetail.texts[i]);
                        countMainDescription++;
                    }
                }

                //Short Description
                for (int i = 0; i < jsonDataContainerDetail.texts.Length; i++)
                {
                    if (jsonDataContainerDetail.texts[i].type.Equals("02", StringComparison.CurrentCultureIgnoreCase))
                    {
                        texts.Add(jsonDataContainerDetail.texts[i]);
                        countShortDescription++;
                    }
                }

                //Long Description
                for (int i = 0; i < jsonDataContainerDetail.texts.Length; i++)
                {
                    if (jsonDataContainerDetail.texts[i].type.Equals("03", StringComparison.CurrentCultureIgnoreCase))
                    {
                        texts.Add(jsonDataContainerDetail.texts[i]);
                        countLongDescription++;
                    }
                }

                //Table of Contents
                for (int i = 0; i < jsonDataContainerDetail.texts.Length; i++)
                {
                    if (jsonDataContainerDetail.texts[i].type.Equals("04", StringComparison.CurrentCultureIgnoreCase))
                    {
                        texts.Add(jsonDataContainerDetail.texts[i]);
                        countTableOfContents++;
                    }
                }

                //Other Texts
                for (int i = 0; i < jsonDataContainerDetail.texts.Length; i++)
                {
                    if (!jsonDataContainerDetail.texts[i].type.Equals("01", StringComparison.CurrentCultureIgnoreCase) &&
                        !jsonDataContainerDetail.texts[i].type.Equals("02", StringComparison.CurrentCultureIgnoreCase) &&
                        !jsonDataContainerDetail.texts[i].type.Equals("03", StringComparison.CurrentCultureIgnoreCase) &&
                        !jsonDataContainerDetail.texts[i].type.Equals("04", StringComparison.CurrentCultureIgnoreCase))
                    {
                        texts.Add(jsonDataContainerDetail.texts[i]);
                        countOtherTexts++;
                    }
                }

                //Init Table Data
                string[,] data = new string[2, texts.Count + 1];
                Font[,] font = new Font[2, texts.Count + 1];
                string[,] hyperlinkValue = new string[2, texts.Count + 1];
                int[] columnWidth = new int[2];
                bool[] columnWordWrap = new bool[2];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 150;
                columnWidth[1] = 630;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = true;

                //Add Header
                data[0, 0] = "";
                data[1, 0] = "Text";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";

                //Add Data to Table
                for (int i = 0; i < texts.Count; i++)
                {
                    data[0, i + 1] = "";
                    font[0, i + 1] = listItemDefaultFontBold;
                    hyperlinkValue[0, i + 1] = "";

                    //Row Header
                    if (countMainDescription != 0 && i == 0)
                    {
                        data[0, i + 1] = "Haupteschreibung";
                    }

                    if (countShortDescription != 0 && i == countMainDescription)
                    {
                        data[0, i + 1] = "Kurzbeschreibung";
                    }

                    if (countLongDescription != 0 && i == countMainDescription + countShortDescription)
                    {
                        data[0, i + 1] = "Vollständige Beschr.";
                    }

                    if (countTableOfContents != 0 && i == countMainDescription + countShortDescription + countLongDescription)
                    {
                        data[0, i + 1] = "Inhalt";
                    }

                    if (countOtherTexts != 0 && i == countMainDescription + countShortDescription + countLongDescription + countTableOfContents)
                    {
                        data[0, i + 1] = "Andere";
                    }

                    //Data
                    data[1, i + 1] = texts[i].value;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[1, i + 1] = "";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[2].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[2].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Prices
            #region prices
            {
                //Variables
                List<Prices> prices = new List<Prices>();

                //Prices
                for (int i = 0; i < jsonDataContainerDetail.prices.Length; i++)
                {
                    prices.Add(jsonDataContainerDetail.prices[i]);
                }

                //Init Table Data
                string[,] data = new string[6, prices.Count + 1];
                Font[,] font = new Font[6, prices.Count + 1];
                string[,] hyperlinkValue = new string[6, prices.Count + 1];
                int[] columnWidth = new int[6];
                bool[] columnWordWrap = new bool[6];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 100;
                columnWidth[1] = 100;
                columnWidth[2] = 150;
                columnWidth[3] = 150;
                columnWidth[4] = 100;
                columnWidth[5] = 100;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;
                columnWordWrap[2] = false;
                columnWordWrap[3] = false;
                columnWordWrap[4] = false;
                columnWordWrap[5] = false;

                //Add Header
                data[0, 0] = "Währung";
                data[1, 0] = "Betrag";
                data[2, 0] = "MwSt.";
                data[3, 0] = "Preisart";
                data[4, 0] = "Gültig ab";
                data[5, 0] = "Gültig bis";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                font[2, 0] = listItemDefaultFontBold;
                font[3, 0] = listItemDefaultFontBold;
                font[4, 0] = listItemDefaultFontBold;
                font[5, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";
                hyperlinkValue[2, 0] = "";
                hyperlinkValue[3, 0] = "";
                hyperlinkValue[4, 0] = "";
                hyperlinkValue[5, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < prices.Count; i++)
                {
                    //Data
                    data[0, i + 1] = String.Format("{0}{1}", prices[i].currency, prices[i].country.Equals("") ? "" : string.Format(" ({0})", prices[i].country));
                    data[1, i + 1] = prices[i].value;
                    data[2, i + 1] = prices[i].taxRate;
                    data[3, i + 1] = prices[i].fixedRetailPrice;
                    data[4, i + 1] = prices[i].validFrom;
                    data[5, i + 1] = prices[i].validUntil;
                    font[0, i + 1] = listItemDefaultFontRegular;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    font[2, i + 1] = listItemDefaultFontRegular;
                    font[3, i + 1] = listItemDefaultFontRegular;
                    font[4, i + 1] = listItemDefaultFontRegular;
                    font[5, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[0, i + 1] = "";
                    hyperlinkValue[1, i + 1] = "";
                    hyperlinkValue[2, i + 1] = "";
                    hyperlinkValue[3, i + 1] = "";
                    hyperlinkValue[4, i + 1] = "";
                    hyperlinkValue[5, i + 1] = "";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[3].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[3].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Identifiers
            #region identifiers
            {
                //Variables
                List<Identifiers> identifiers = new List<Identifiers>();

                //Identifiers
                for (int i = 0; i < jsonDataContainerDetail.identifiers.Length; i++)
                {
                    identifiers.Add(jsonDataContainerDetail.identifiers[i]);
                }

                //Init Table Data
                string[,] data = new string[2, identifiers.Count + 1];
                Font[,] font = new Font[2, identifiers.Count + 1];
                string[,] hyperlinkValue = new string[2, identifiers.Count + 1];
                int[] columnWidth = new int[2];
                bool[] columnWordWrap = new bool[2];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 100;
                columnWidth[1] = 300;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;

                //Add Header
                data[0, 0] = "Bezeichner";
                data[1, 0] = "Nummer";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < identifiers.Count; i++)
                {
                    //Data
                    data[0, i + 1] = codeList_005_identifier.getCodeListValue(identifiers[i].type);
                    data[1, i + 1] = identifiers[i].value;
                    font[0, i + 1] = listItemDefaultFontRegular;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[0, i + 1] = "";
                    hyperlinkValue[1, i + 1] = "";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[4].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[4].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Productstatus
            #region productstatus
            {
                //Variables
                List<string> productstatusType = new List<string>();
                List<string> productstatusValue = new List<string>();

                //Productstatus
                productstatusType.Add("Erscheinungstermin");
                productstatusType.Add("Lieferbarkeitsstatus");
                productstatusType.Add("Publikationsstatus");
                productstatusType.Add("Zugang");
                productstatusType.Add("Letzte Aktualisierung");
                productstatusType.Add("Letzte Preisänderung");

                productstatusValue.Add(jsonDataContainerDetail.publicationDate);
                productstatusValue.Add(codeList_065_availability.getCodeListValue(jsonDataContainerDetail.availability));
                productstatusValue.Add(codeList_064_publishingStatus.getCodeListValue(jsonDataContainerDetail.publishingStatus));
                productstatusValue.Add(jsonDataContainerDetail.createdDate);
                productstatusValue.Add(jsonDataContainerDetail.lastModifiedDate);
                productstatusValue.Add("TODO");

                //Init Table Data
                string[,] data = new string[2, productstatusType.Count + 1];
                Font[,] font = new Font[2, productstatusType.Count + 1];
                string[,] hyperlinkValue = new string[2, productstatusType.Count + 1];
                int[] columnWidth = new int[2];
                bool[] columnWordWrap = new bool[2];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 150;
                columnWidth[1] = 600;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;

                //Add Header
                data[0, 0] = "";
                data[1, 0] = "Datum / Status";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < productstatusType.Count; i++)
                {
                    //Data
                    data[0, i + 1] = productstatusType[i];
                    data[1, i + 1] = productstatusValue[i];
                    font[0, i + 1] = listItemDefaultFontBold;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[0, i + 1] = "";
                    hyperlinkValue[1, i + 1] = "";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[5].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[5].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Classifications
            #region classifications
            {
                //Variables
                List<Classifications> classifications = jsonDataContainerDetail.classifications.ToList<Classifications>();
                string previousItem = null;
                
                //Sort
                classifications = classifications.OrderBy(o => o.type).ToList();

                //Init Table Data
                string[,] data = new string[5, classifications.Count + 1];
                Font[,] font = new Font[5, classifications.Count + 1];
                string[,] hyperlinkValue = new string[5, classifications.Count + 1];
                int[] columnWidth = new int[5];
                bool[] columnWordWrap = new bool[5];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 200;
                columnWidth[1] = 100;
                columnWidth[2] = 300;
                columnWidth[3] = 100;
                columnWidth[4] = 50;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;
                columnWordWrap[2] = false;
                columnWordWrap[3] = false;
                columnWordWrap[4] = false;

                //Add Header
                data[0, 0] = "";
                data[1, 0] = "Code";
                data[2, 0] = "Text";
                data[3, 0] = "Quelle";
                data[4, 0] = "Version";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                font[2, 0] = listItemDefaultFontBold;
                font[3, 0] = listItemDefaultFontBold;
                font[4, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";
                hyperlinkValue[2, 0] = "";
                hyperlinkValue[3, 0] = "";
                hyperlinkValue[4, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < classifications.Count; i++)
                {
                    data[0, i + 1] = "";
                    font[0, i + 1] = listItemDefaultFontBold;
                    hyperlinkValue[0, i + 1] = "";

                    //Row Header
                    if (!classifications[i].type.Equals(previousItem, StringComparison.CurrentCultureIgnoreCase))
                    {
                        previousItem = classifications[i].type;

                        data[0, i + 1] = codeList_027_subjectIdentifier.getCodeListValue(classifications[i].type);
                    }

                    //Data
                    data[1, i + 1] = classifications[i].code;
                    data[2, i + 1] = classifications[i].value;
                    data[3, i + 1] = classifications[i].sourceName;
                    data[4, i + 1] = classifications[i].systematicVersion;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    font[2, i + 1] = listItemDefaultFontRegular;
                    font[3, i + 1] = listItemDefaultFontRegular;
                    font[4, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[1, i + 1] = "";
                    hyperlinkValue[2, i + 1] = classifications[i].type.Equals("20") ? "SW" : "";
                    hyperlinkValue[3, i + 1] = "";
                    hyperlinkValue[4, i + 1] = "";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[6].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[6].Controls.Add(forms[i]);
                }
            }
            #endregion

            //MediaFiles
            #region mediaFiles
            {
                //Variables
                if(jsonDataContainerMedia != null)
                {
                    List<JsonDataContainerMedia> mediaFiles = jsonDataContainerMedia.ToList<JsonDataContainerMedia>();

                    //Sort
                    mediaFiles = mediaFiles.OrderBy(o => o.type).ToList();

                    //Init Table Data
                    string[,] data = new string[3, mediaFiles.Count + 1];
                    Font[,] font = new Font[3, mediaFiles.Count + 1];
                    string[,] hyperlinkValue = new string[3, mediaFiles.Count + 1];
                    int[] columnWidth = new int[3];
                    bool[] columnWordWrap = new bool[3];
                    int spacingX;
                    int spacingY;
                    int offsetX;
                    int offsetY;

                    //Config Table
                    columnWidth[0] = 300;
                    columnWidth[1] = 100;
                    columnWidth[2] = 400;

                    spacingX = 7;
                    spacingY = 5;
                    offsetX = 8;
                    offsetY = 8;

                    columnWordWrap[0] = false;
                    columnWordWrap[1] = false;
                    columnWordWrap[2] = true;

                    //Add Header
                    data[0, 0] = "Typ";
                    data[1, 0] = "Nummer";
                    data[2, 0] = "Datei herunterladen";
                    font[0, 0] = listItemDefaultFontBold;
                    font[1, 0] = listItemDefaultFontBold;
                    font[2, 0] = listItemDefaultFontBold;
                    hyperlinkValue[0, 0] = "";
                    hyperlinkValue[1, 0] = "";
                    hyperlinkValue[2, 0] = "";

                    //Write Data to Detail Tab
                    for (int i = 0; i < mediaFiles.Count; i++)
                    {
                        //Data
                        data[0, i + 1] = codeList_MediaData.getCodeListValue(mediaFiles[i].type);
                        data[1, i + 1] = mediaFiles[i].sequenceNumber;
                        data[2, i + 1] = mediaFiles[i].url;
                        font[0, i + 1] = listItemDefaultFontRegular;
                        font[1, i + 1] = listItemDefaultFontRegular;
                        font[2, i + 1] = listItemDefaultFontRegular;
                        hyperlinkValue[0, i + 1] = "";
                        hyperlinkValue[1, i + 1] = "";
                        hyperlinkValue[2, i + 1] = "DOWNLOAD";
                    }

                    //Add Forms to TabPage
                    List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                    tabControl_detail.TabPages[7].Controls.Clear();

                    for (int i = 0; i < forms.Count; i++)
                    {
                        tabControl_detail.TabPages[7].Controls.Add(forms[i]);
                    }
                }
            }
            #endregion

            //Collections
            #region collections

            {
                //Variables
                List<Collections> collections = jsonDataContainerDetail.collections.ToList<Collections>();

                //Init Table Data
                string[,] data = new string[4, collections.Count + 1];
                Font[,] font = new Font[4, collections.Count + 1];
                string[,] hyperlinkValue = new string[4, collections.Count + 1];
                int[] columnWidth = new int[4];
                bool[] columnWordWrap = new bool[4];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 250;
                columnWidth[1] = 100;
                columnWidth[2] = 250;
                columnWidth[3] = 150;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;
                columnWordWrap[2] = false;
                columnWordWrap[3] = true;

                //Add Header
                data[0, 0] = "Titel";
                data[1, 0] = "Position";
                data[2, 0] = "Verlag";
                data[3, 0] = "Nummer(n)";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                font[2, 0] = listItemDefaultFontBold;
                font[3, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";
                hyperlinkValue[2, 0] = "";
                hyperlinkValue[3, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < collections.Count; i++)
                {
                    //Data
                    data[0, i + 1] = collections[i].master.title;
                    data[1, i + 1] = collections[i].levelSequence;
                    data[2, i + 1] = collections[i].master.publisherNameVlb;
                    data[3, i + 1] = "";

                    for(int j = 0; j < collections[i].master.identifiers.Length; j++)
                    {
                        data[3, i + 1] += collections[i].master.identifiers[j].value;

                        if(j != collections[i].master.identifiers.Length - 1)
                        {
                            data[3, i + 1] += "\n";
                        }
                    }

                    font[0, i + 1] = listItemDefaultFontRegular;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    font[2, i + 1] = listItemDefaultFontRegular;
                    font[3, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[0, i + 1] = "";
                    hyperlinkValue[1, i + 1] = "";
                    hyperlinkValue[2, i + 1] = "VL";
                    hyperlinkValue[3, i + 1] = "RH";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[8].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[8].Controls.Add(forms[i]);
                }
            }
            #endregion

            //Wholesalers
            #region wholesalers

            {
                //Variables
                List<Wholesalers> wholesalers = jsonDataContainerDetail.wholesalers.ToList<Wholesalers>();

                //Init Table Data
                string[,] data = new string[4, wholesalers.Count + 1];
                Font[,] font = new Font[4, wholesalers.Count + 1];
                string[,] hyperlinkValue = new string[4, wholesalers.Count + 1];
                int[] columnWidth = new int[4];
                bool[] columnWordWrap = new bool[4];
                int spacingX;
                int spacingY;
                int offsetX;
                int offsetY;

                //Config Table
                columnWidth[0] = 100;
                columnWidth[1] = 100;
                columnWidth[2] = 200;
                columnWidth[3] = 100;

                spacingX = 7;
                spacingY = 5;
                offsetX = 8;
                offsetY = 8;

                columnWordWrap[0] = false;
                columnWordWrap[1] = false;
                columnWordWrap[2] = false;
                columnWordWrap[3] = false;

                //Add Header
                data[0, 0] = "ISBN";
                data[1, 0] = "SIGL";
                data[2, 0] = "Bestellnummer";
                data[3, 0] = "Quelle";
                font[0, 0] = listItemDefaultFontBold;
                font[1, 0] = listItemDefaultFontBold;
                font[2, 0] = listItemDefaultFontBold;
                font[3, 0] = listItemDefaultFontBold;
                hyperlinkValue[0, 0] = "";
                hyperlinkValue[1, 0] = "";
                hyperlinkValue[2, 0] = "";
                hyperlinkValue[3, 0] = "";

                //Write Data to Detail Tab
                for (int i = 0; i < wholesalers.Count; i++)
                {
                    //Data
                    data[0, i + 1] = wholesalers[i].isbn;
                    data[1, i + 1] = wholesalers[i].sigl;
                    data[2, i + 1] = wholesalers[i].orderNumber;
                    data[3, i + 1] = wholesalers[i].recordSource;
                    font[0, i + 1] = listItemDefaultFontRegular;
                    font[1, i + 1] = listItemDefaultFontRegular;
                    font[2, i + 1] = listItemDefaultFontRegular;
                    font[3, i + 1] = listItemDefaultFontRegular;
                    hyperlinkValue[0, i + 1] = "";
                    hyperlinkValue[1, i + 1] = "";
                    hyperlinkValue[2, i + 1] = "";
                    hyperlinkValue[3, i + 1] = "";
                }

                //Add Forms to TabPage
                List<Control> forms = arrangeDataInTable(data, font, hyperlinkValue, columnWidth, columnWordWrap, spacingX, spacingY, offsetX, offsetY);

                tabControl_detail.TabPages[9].Controls.Clear();

                for (int i = 0; i < forms.Count; i++)
                {
                    tabControl_detail.TabPages[9].Controls.Add(forms[i]);
                }
            }
            #endregion



            //Set Publisher Detail View
            getPublisherDetail(currentPublisherMVBID);



            //Switch to Detail View
            tabControl.SelectTab(1);

            //Save ISBN for QRCode Generator
            currentlyShownISBN = isbn.Replace("-", "");
            currentlyShownID = jsonDataContainerDetail.id;

            //Get Cover
            pictureBox_detail_cover.Image = Image.FromFile(Application.StartupPath + "/bin/cover_loading.png");
            cover = null;

            BackgroundWorker backgroundWorker = new BackgroundWorker();
            backgroundWorker.DoWork += getCoverAsyncDoWork;
            backgroundWorker.RunWorkerCompleted += getCoverAsyncCompleted;

            backgroundWorker.RunWorkerAsync();
        }



        /// <summary>
        /// Index-Change Event Handler - sorts the current Result List by the selected Field.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void comboBox_search_sortingField_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (comboBox_search_sortingField.Text)
            {
                case "Relevanz":
                    sortField = "score";
                    break;

                case "ISBN":
                    sortField = "isbn";
                    break;

                case "Titel":
                    sortField = "title";
                    break;

                case "Autor":
                    sortField = "author";
                    break;

                case "Verlag":
                    sortField = "publisher";
                    break;

                case "Preis (EUR D)":
                    sortField = "price";
                    break;

                case "Produkttyp":
                    sortField = "productType";
                    break;
            }

            if (atLeastOneSearchPerformed)
            {
                page = 1;
                performSearch();
            }
        }

        /// <summary>
        /// Index-Change Event Handler - sorts the current Result List in a different Direction (ascending, descending).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void comboBox_search_sortingDirection_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (comboBox_search_sortingDirection.Text.Equals("Aufsteigend"))
            {
                sortDirection = "asc";
            }

            if (comboBox_search_sortingDirection.Text.Equals("Absteigend"))
            {
                sortDirection = "desc";
            }

            if (atLeastOneSearchPerformed)
            {
                page = 1;
                performSearch();
            }
        }

        /// <summary>
        /// Index-Change Event Handler - shows a specified Amound of Results.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void comboBox_search_itemsPerPage_SelectedIndexChanged(object sender, EventArgs e)
        {
            amountOfTitlesShown = int.Parse(comboBox_search_itemsPerPage.Text);

            if (atLeastOneSearchPerformed)
            {
                page = 1;
                performSearch();
            }
        }



        /// <summary>
        /// Performs a new Search with page = 1.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_first_Click(object sender, EventArgs e)
        {
            if(page > 1)
            {
                page = 1;
                performSearch();
            }

            setPageButtonStatus();
        }

        /// <summary>
        /// Performs a new Search with page -= 1.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_previous_Click(object sender, EventArgs e)
        {
            if(page > 1)
            {
                page--;
                performSearch();
            }

            setPageButtonStatus();
        }

        /// <summary>
        /// Performs a new Search with page += 1.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_next_Click(object sender, EventArgs e)
        {
            if(page < lastPage)
            {
                page++;
                performSearch();
            }

            setPageButtonStatus();
        }

        /// <summary>
        /// Performs a new Search with page = lastPage.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_last_Click(object sender, EventArgs e)
        {

            if (page < lastPage)
            {
                page = lastPage;
                performSearch();
            }

            setPageButtonStatus();
        }



        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByISBN_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "identifier", waterMarkTextBox_search_searchByISBN, "ISBN / EAN / GTIN");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByTitle_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "title", waterMarkTextBox_search_searchByTitle, "Titel");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByAuthor_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "author", waterMarkTextBox_search_searchByAuthor, "Autor");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByPublisher_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "publisher", waterMarkTextBox_search_searchByPublisher, "Verlag");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchBySubject_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "keyword", waterMarkTextBox_search_searchBySubject, "Schlagwort");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchBySet_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "set", waterMarkTextBox_search_searchBySet, "Reihe");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// Index-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByCollection_Click(object sender, EventArgs e)
        {
            IndexFrame indexFrame = new IndexFrame(statefullAccessToken, "collection", waterMarkTextBox_search_searchByCollection, "Hierarchie");
            indexFrame.ShowDialog();
        }

        /// <summary>
        /// AutoComplete-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByProductForm_Click(object sender, EventArgs e)
        {
            AutoCompleteForm autoCompleteFrame = new AutoCompleteForm(waterMarkTextBox_search_searchByProductForm, codeList_007_productform, "Produktform");
            autoCompleteFrame.ShowDialog();
        }

        /// <summary>
        /// AutoComplete-Call.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_search_searchByLanguage_Click(object sender, EventArgs e)
        {
            AutoCompleteForm autoCompleteFrame = new AutoCompleteForm(waterMarkTextBox_search_searchByLanguage, codeList_074_language, "Sprache");
            autoCompleteFrame.ShowDialog();
        }



        /// <summary>
        /// Saves the current Debug Log into a Textfile.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void button_debug_saveDebugText_Click(object sender, EventArgs e)
        {
            //Show Dialog
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "Textdatei|*.txt";
            saveFileDialog.Title = "Debug-Informationen in eine Textdatei exportieren";
            saveFileDialog.AddExtension = true;
            saveFileDialog.OverwritePrompt = true;

            //Write Output
            if(saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    //Variables
                    FileStream fileStream = new FileStream(saveFileDialog.FileName, FileMode.Create);
                    StreamWriter streamWriter = new StreamWriter(fileStream);

                    //Write Text
                    streamWriter.Write(richTextBox_debug_rawData.Text);

                    //Close Streams
                    streamWriter.Flush();
                    streamWriter.Close();
                    fileStream.Close();
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Fehler");
                }
            }
        }



        /// <summary>
        /// Changes some MenuItems when changing the tabControls Index.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void tabControl_selectedIndexChange(object sender, EventArgs e)
        {
            switch (tabControl.SelectedTab.Name)
            {
                case "tabPage_search":
                    ToolStripMenuItem_orderSelection.Enabled = true;
                    ToolStripMenuItem_exportSelection.Enabled = true;
                    ToolStripMenuItem_exportSelectionAsQRCode.Enabled = true;

                    ToolStripMenuItem_orderTitle.Enabled = false;
                    ToolStripMenuItem_exportTitle.Enabled = false;
                    ToolStripMenuItem_exportTitleAsQRCode.Enabled = false;

                    ToolStripMenuItem_viewTitleOnVLB.Enabled = false;
                    ToolStripMenuItem_viewTitleOnBuchhandelDe.Enabled = false;
                    break;

                case "tabPage_details":
                    ToolStripMenuItem_orderSelection.Enabled = false;
                    ToolStripMenuItem_exportSelection.Enabled = false;
                    ToolStripMenuItem_exportSelectionAsQRCode.Enabled = false;

                    ToolStripMenuItem_orderTitle.Enabled = true;
                    ToolStripMenuItem_exportTitle.Enabled = true;
                    ToolStripMenuItem_exportTitleAsQRCode.Enabled = true;

                    ToolStripMenuItem_viewTitleOnVLB.Enabled = true;
                    ToolStripMenuItem_viewTitleOnBuchhandelDe.Enabled = true;
                    break;

                case "tabPage_publisher":
                    ToolStripMenuItem_orderSelection.Enabled = false;
                    ToolStripMenuItem_exportSelection.Enabled = false;
                    ToolStripMenuItem_exportSelectionAsQRCode.Enabled = false;

                    ToolStripMenuItem_orderTitle.Enabled = false;
                    ToolStripMenuItem_exportTitle.Enabled = false;
                    ToolStripMenuItem_exportTitleAsQRCode.Enabled = false;

                    ToolStripMenuItem_viewTitleOnVLB.Enabled = false;
                    ToolStripMenuItem_viewTitleOnBuchhandelDe.Enabled = false;
                    break;

                case "tabPage_debug":
                    ToolStripMenuItem_orderSelection.Enabled = false;
                    ToolStripMenuItem_exportSelection.Enabled = false;
                    ToolStripMenuItem_exportSelectionAsQRCode.Enabled = false;

                    ToolStripMenuItem_orderTitle.Enabled = false;
                    ToolStripMenuItem_exportTitle.Enabled = false;
                    ToolStripMenuItem_exportTitleAsQRCode.Enabled = false;

                    ToolStripMenuItem_viewTitleOnVLB.Enabled = false;
                    ToolStripMenuItem_viewTitleOnBuchhandelDe.Enabled = false;
                    break;

            }
        }



        /// <summary>
        /// Opens the Login Data Frame.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_accessData_Click(object sender, EventArgs e)
        {
            LoginDataFrame loginDataFrame = new LoginDataFrame(RBSPassword);
            loginDataFrame.ShowDialog();
        }



        /// <summary>
        /// Closes the Application.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_close_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }



        /// <summary>
        /// Creates a new Order File.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_orderSelection_Click(object sender, EventArgs e)
        {
            //Variables
            DataGridViewSelectedRowCollection rows = dataGridView_search_results.SelectedRows;
            List<string> ISBNs = new List<string>();

            //Get ISBNs to export
            for (int i = 0; i < rows.Count; i++)
            {
                ISBNs.Add(rows[i].Cells[0].Value.ToString().Replace("-", "").Trim());
            }

            //No Titles selected
            if (ISBNs.Count == 0)
            {
                MessageBox.Show("Sie haben keine Titel in der Trefferliste selektiert", "Fehler");
                return;
            }

            //Titles without ISBN selected
            if (ISBNs.Contains(""))
            {
                MessageBox.Show("Titel ohne ISBN können nicht exportiert werden.\nBitte entfernen Sie diese aus ihrer Auswahl.", "Fehler");
                return;
            }

            //Titles with malformed ISBN selected
            for (int i = 0; i < ISBNs.Count; i++)
            {
                if (ISBNs[i].Length != 13)
                {
                    MessageBox.Show("Nur Titel mit 13-stelliger ISBN können exportiert werden.\nBitte entfernen Sie Titel mit ungültigem Bezeichner aus ihrer Auswahl.", "Fehler");
                    return;
                }
            }

            //Export Titles to File
            export_order(ISBNs);
        }



        /// <summary>
        /// Exports the selected Titles as .txt Textfile.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_exportSelection_Click(object sender, EventArgs e)
        {
            //Variables
            DataGridViewSelectedRowCollection rows = dataGridView_search_results.SelectedRows;
            List<string> ISBNs = new List<string>();

            //Get ISBNs to export
            for (int i = 0; i < rows.Count; i++)
            {
                ISBNs.Add(rows[i].Cells[0].Value.ToString().Replace("-", "").Trim());
            }

            //No Titles selected
            if (ISBNs.Count == 0)
            {
                MessageBox.Show("Sie haben keine Titel in der Trefferliste selektiert", "Fehler");
                return;
            }

            //Titles without ISBN selected
            if (ISBNs.Contains(""))
            {
                MessageBox.Show("Titel ohne ISBN können nicht exportiert werden.\nBitte entfernen Sie diese aus ihrer Auswahl.", "Fehler");
                return;
            }

            //Titles with malformed ISBN selected
            for (int i = 0; i < ISBNs.Count; i++)
            {
                if (ISBNs[i].Length != 13)
                {
                    MessageBox.Show("Nur Titel mit 13-stelliger ISBN können exportiert werden.\nBitte entfernen Sie Titel mit ungültigem Bezeichner aus ihrer Auswahl.", "Fehler");
                    return;
                }
            }

            //Export Titles to File
            export_asTXT(ISBNs);
        }



        /// <summary>
        /// Exports the selected Titles as buchhandel.de QR-Codes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_exportSelectionAsQRCode_Click(object sender, EventArgs e)
        {
            //Variables
            DataGridViewSelectedRowCollection rows = dataGridView_search_results.SelectedRows;
            List<string> ISBNs = new List<string>();

            //Get ISBNs to export
            for (int i = 0; i < rows.Count; i++)
            {
                ISBNs.Add(rows[i].Cells[0].Value.ToString().Replace("-", "").Trim());
            }

            //No Titles selected
            if (ISBNs.Count == 0)
            {
                MessageBox.Show("Sie haben keine Titel in der Trefferliste selektiert", "Fehler");
                return;
            }

            //Titles without ISBN selected
            if (ISBNs.Contains(""))
            {
                MessageBox.Show("Titel ohne ISBN können nicht exportiert werden.\nBitte entfernen Sie diese aus ihrer Auswahl.", "Fehler");
                return;
            }

            //Titles with malformed ISBN selected
            for (int i = 0; i < ISBNs.Count; i++)
            {
                if (ISBNs[i].Length != 13)
                {
                    MessageBox.Show("Nur Titel mit 13-stelliger ISBN können exportiert werden.\nBitte entfernen Sie Titel mit ungültigem Bezeichner aus ihrer Auswahl.", "Fehler");
                    return;
                }
            }

            //Show Folder Browser Dialog
            FolderBrowserDialog folderBrowserDialog = new FolderBrowserDialog();
            folderBrowserDialog.RootFolder = Environment.SpecialFolder.Desktop;

            if(folderBrowserDialog.ShowDialog() == DialogResult.OK)
            {
                //Generate Codes for all Titles
                for (int i = 0; i < ISBNs.Count; i++)
                {
                    //Save File
                    if (!export_asQRCode(ISBNs[i], string.Format("{0}/{1}.png", folderBrowserDialog.SelectedPath, ISBNs[i])))
                    {
                        return;
                    }
                }
            }

        }



        /// <summary>
        /// Creates a new Order File.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_orderTitle_Click(object sender, EventArgs e)
        {
            //No Title selected
            if (currentlyShownISBN.Equals(""))
            {
                MessageBox.Show("Sie haben bislang keinen Titel in der Titeldetailansicht angezeigt.\n\nNur ein aktuell in der Titeldetailansicht angezeigter Titel kann bestellt werden.", "Info");
                return;
            }

            //Export Title to File
            List<string> ISBN = new List<string>();
            ISBN.Add(currentlyShownISBN);
            export_order(ISBN);
        }



        /// <summary>
        /// Exports the current Title as .txt Textfile.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_exportTitle_Click(object sender, EventArgs e)
        {
            //No Title selected
            if (currentlyShownISBN.Equals(""))
            {
                MessageBox.Show("Sie haben bislang keinen Titel in der Titeldetailansicht angezeigt.\n\nNur ein aktuell in der Titeldetailansicht angezeigter Titel kann bestellt werden.", "Info");
                return;
            }

            //Export Title to File
            List<string> ISBN = new List<string>();
            ISBN.Add(currentlyShownISBN);
            export_asTXT(ISBN);
        }



        /// <summary>
        /// Generates a buchhandel.de QR-Code and saves it to the user defined Location.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_exportTitleAsQRCode_Click(object sender, EventArgs e)
        {
            //No Title selected
            if (currentlyShownISBN.Equals(""))
            {
                MessageBox.Show("Sie haben bislang keinen Titel in der Titeldetailansicht angezeigt.\n\nNur ein aktuell in der Titeldetailansicht angezeigter Titel kann als QR-Code exportiert werden.", "Info");
                return;
            }

            //Init Savefile Dialog
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "png | *.png";
            saveFileDialog.InitialDirectory = Application.StartupPath;
            saveFileDialog.FileName = currentlyShownISBN + ".png";

            //Save File
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                if(!export_asQRCode(currentlyShownISBN, saveFileDialog.FileName))
                {
                    return;
                }
            }
        }



        /// <summary>
        /// Opens up the prefered Browsers and searches the Title on VLB.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_viewTitleOnVLB_Click(object sender, EventArgs e)
        {
            //No Title selected
            if (currentlyShownISBN.Equals(""))
            {
                MessageBox.Show("Sie haben bislang keinen Titel in der Titeldetailansicht angezeigt.", "Info");
                return;
            }

            //Open prefered Browser
            try
            {
                System.Diagnostics.Process.Start(@"https://prod.vlb.de/#search/quicksearch/startQuickSearch/" + currentlyShownISBN);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
            }
        }



        /// <summary>
        /// Opens up the prefered Browsers and searches the Title on buchhandel.de.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_viewTitleOnBuchhandelDe_Click(object sender, EventArgs e)
        {
            //No Title selected
            if (currentlyShownISBN.Equals(""))
            {
                MessageBox.Show("Sie haben bislang keinen Titel in der Titeldetailansicht angezeigt.", "Info");
                return;
            }

            //Open prefered Browser
            try
            {
                System.Diagnostics.Process.Start(@"https://www.buchhandel.de/suche?q=" + currentlyShownISBN + "&wgi=");
            }
            catch (Exception ex)
            {
                MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
            }
        }



        /// <summary>
        /// Opens the User Manual.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ToolStripMenuItem_manual_Click(object sender, EventArgs e)
        {
            try
            {
                System.Diagnostics.Process.Start(Application.StartupPath + "/bin/RestBookServiceManual.pdf");
            }
            catch (Exception ex)
            {
                MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
            }
        }



        /// <summary>
        /// Handles all Clicks at LinkLabels, performing a new Search with the given Link Parameters.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void linkLabel_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            if (e.Link.Name.Trim().Equals("IS*", StringComparison.CurrentCultureIgnoreCase))
            {
                searchParam_quickSearch = "";
                searchParam_searchByISBN = ((LinkLabel)sender).Text + "*";
                searchParam_searchByTitle = "";
                searchParam_searchByAuthor = "";
                searchParam_searchByPublisher = "";
                searchParam_searchBySubject = "";
                searchParam_searchBySet = "";
                searchParam_searchByCollection = "";
                searchParam_searchByProductForm = "";
                searchParam_searchByLanguage = "";
                searchParam_searchByPriceFrom = "";
                searchParam_searchByPriceTo = "";
                searchParam_searchByPublicationDateFrom = "";
                searchParam_searchByPublicationDateTo = "";

                //Perform Search
                page = 1;
                performSearch();
                setPageButtonStatus();

                //Switch to Search View
                tabControl.SelectTab(0);
            }

            if (e.Link.Name.Trim().Equals("AU", StringComparison.CurrentCultureIgnoreCase))
            {
                searchParam_quickSearch = "";
                searchParam_searchByISBN = "";
                searchParam_searchByTitle = "";
                searchParam_searchByAuthor = ((LinkLabel)sender).Text;
                searchParam_searchByPublisher = "";
                searchParam_searchBySubject = "";
                searchParam_searchBySet = "";
                searchParam_searchByCollection = "";
                searchParam_searchByProductForm = "";
                searchParam_searchByLanguage = "";
                searchParam_searchByPriceFrom = "";
                searchParam_searchByPriceTo = "";
                searchParam_searchByPublicationDateFrom = "";
                searchParam_searchByPublicationDateTo = "";

                //Perform Search
                page = 1;
                performSearch();
                setPageButtonStatus();

                //Switch to Search View
                tabControl.SelectTab(0);
            }

            if (e.Link.Name.Trim().Equals("VL", StringComparison.CurrentCultureIgnoreCase))
            {
                searchParam_quickSearch = "";
                searchParam_searchByISBN = "";
                searchParam_searchByTitle = "";
                searchParam_searchByAuthor = "";
                searchParam_searchByPublisher = ((LinkLabel)sender).Text;
                searchParam_searchBySubject = "";
                searchParam_searchBySet = "";
                searchParam_searchByCollection = "";
                searchParam_searchByProductForm = "";
                searchParam_searchByLanguage = "";
                searchParam_searchByPriceFrom = "";
                searchParam_searchByPriceTo = "";
                searchParam_searchByPublicationDateFrom = "";
                searchParam_searchByPublicationDateTo = "";

                //Perform Search
                page = 1;
                performSearch();
                setPageButtonStatus();

                //Switch to Search View
                tabControl.SelectTab(0);
            }

            if (e.Link.Name.Trim().Equals("SW", StringComparison.CurrentCultureIgnoreCase))
            {
                searchParam_quickSearch = "";
                searchParam_searchByISBN = "";
                searchParam_searchByTitle = "";
                searchParam_searchByAuthor = "";
                searchParam_searchByPublisher = "";
                searchParam_searchBySubject = ((LinkLabel)sender).Text;
                searchParam_searchBySet = "";
                searchParam_searchByCollection = "";
                searchParam_searchByProductForm = "";
                searchParam_searchByLanguage = "";
                searchParam_searchByPriceFrom = "";
                searchParam_searchByPriceTo = "";
                searchParam_searchByPublicationDateFrom = "";
                searchParam_searchByPublicationDateTo = "";

                //Perform Search
                page = 1;
                performSearch();
                setPageButtonStatus();

                //Switch to Search View
                tabControl.SelectTab(0);
            }

            if (e.Link.Name.Trim().Equals("RH", StringComparison.CurrentCultureIgnoreCase))
            {
                searchParam_quickSearch = "";
                searchParam_searchByISBN = "";
                searchParam_searchByTitle = "";
                searchParam_searchByAuthor = "";
                searchParam_searchByPublisher = "";
                searchParam_searchBySubject = "";
                searchParam_searchBySet = ((LinkLabel)sender).Text;
                searchParam_searchByCollection = "";
                searchParam_searchByProductForm = "";
                searchParam_searchByLanguage = "";
                searchParam_searchByPriceFrom = "";
                searchParam_searchByPriceTo = "";
                searchParam_searchByPublicationDateFrom = "";
                searchParam_searchByPublicationDateTo = "";

                //Perform Search
                page = 1;
                performSearch();
                setPageButtonStatus();

                //Switch to Search View
                tabControl.SelectTab(0);
            }

            if (e.Link.Name.Trim().Equals("URL", StringComparison.CurrentCultureIgnoreCase))
            {
                try
                {
                    System.Diagnostics.Process.Start(((LinkLabel)sender).Text);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                }
            }

            if (e.Link.Name.Trim().Equals("Email", StringComparison.CurrentCultureIgnoreCase))
            {
                try
                {
                    System.Diagnostics.Process.Start(@"mailto:" + ((LinkLabel)sender).Text);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                }
            }

            if (e.Link.Name.Trim().Equals("DOWNLOAD", StringComparison.CurrentCultureIgnoreCase))
            {
                //Variables
                RestClient client;
                HttpWebResponse httpWebResponse;
                Stream stream;
                string contentTypeHeader = "";
                string fileFormat = "dat";
                CultureInfo cultureInfo = CultureInfo.CurrentCulture;

                //Configure REST Request
                client = new RestClient();
                client.EndPoint = ((LinkLabel)sender).Text;
                client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
                client.ContentTypeString = "application/json";
                client.Method = HttpVerb.GET;

                //Send Request and get Image, if possible
                try
                {
                    //Init Stream, get Header Information about File Format
                    httpWebResponse = client.MakeRequestReturnResponse("");
                    stream = httpWebResponse.GetResponseStream();

                    //Guess FileFormat
                    contentTypeHeader = httpWebResponse.Headers["Content-Type"];

                    if (cultureInfo.CompareInfo.IndexOf(contentTypeHeader, "image/jpeg", CompareOptions.IgnoreCase) >= 0)
                    {
                        fileFormat = "jpg";
                    }

                    if (cultureInfo.CompareInfo.IndexOf(contentTypeHeader, "application/pdf", CompareOptions.IgnoreCase) >= 0)
                    {
                        fileFormat = "pdf";
                    }

                    if (cultureInfo.CompareInfo.IndexOf(contentTypeHeader, "video/x-msvideo", CompareOptions.IgnoreCase) >= 0)
                    {
                        fileFormat = "avi";
                    }

                    if (cultureInfo.CompareInfo.IndexOf(contentTypeHeader, "audio/mpeg", CompareOptions.IgnoreCase) >= 0)
                    {
                        fileFormat = "mp3";
                    }

                    if (cultureInfo.CompareInfo.IndexOf(contentTypeHeader, "application/epub+zip", CompareOptions.IgnoreCase) >= 0)
                    {
                        fileFormat = "zip";
                    }

                    //Init Savefile Dialog
                    SaveFileDialog saveFileDialog = new SaveFileDialog();
                    saveFileDialog.Filter = string.Format("{0} | *.{0}", fileFormat);
                    saveFileDialog.InitialDirectory = Application.StartupPath;
                    saveFileDialog.FileName = currentlyShownISBN + "." + fileFormat;

                    if(saveFileDialog.ShowDialog() == DialogResult.OK)
                    {
                        //Write File to Folder
                        FileStream fileStream = File.Create(saveFileDialog.FileName);
                        stream.CopyTo(fileStream);
                        fileStream.Flush();
                        fileStream.Close();
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                }

                richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
            }
        }



        /// <summary>
        /// Listener for the onClosing Event. Performs a Logout if necessary.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainFrame_FormClosing(Object sender, FormClosingEventArgs e)
        {
            //Form closing, logout
            if (!statefullAccessToken.Equals(""))
            {
                MessageBox.Show("Das Programm loggt sich aus. Dies kann bis zu einer Minute in Anspruch nehmen.", "Info");

                RestClient client = new RestClient();
                client.EndPoint = @"http://api.vlb.de/api/v1/logout";
                client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
                client.ContentTypeString = "application/json";
                client.Method = HttpVerb.GET;

                //Send Request
                try
                {
                    string notUsed = "";
                    client.MakeRequest("", ref notUsed);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Fehler");
                    return;
                }

                MessageBox.Show("Sie wurden vom VLB abgemeldet.", "Info");
            }
        }



        /// <summary>
        /// Sets the Page Buttons to enabled or disabled, depending on the currently shown Page.
        /// </summary>
        private void setPageButtonStatus()
        {
            if(page <= 1)
            {
                button_search_first.Enabled = false;
                button_search_previous.Enabled = false;
            }
            else
            {
                button_search_first.Enabled = true;
                button_search_previous.Enabled = true;
            }

            if (page >= lastPage)
            {
                button_search_next.Enabled = false;
                button_search_last.Enabled = false;
            }
            else
            {
                button_search_next.Enabled = true;
                button_search_last.Enabled = true;
            }

            textBox_search_currentPage.Text = string.Format("{0} / {1}", page, lastPage);
        }



        /// <summary>
        /// Performs a new Search with the currently set Search Parameter Buffer.
        /// </summary>
        private void performSearch()
        {
            //Variables
            RestClient client;
            HttpWebResponse json;
            List<string> searchStrings = new List<string>();
            string request;
            string responseAsString = null;

            //Get Search Strings
            if (!searchParam_quickSearch.Equals(""))
            {
                searchStrings.Add(searchParam_quickSearch);
            }

            if (!searchParam_searchByISBN.Equals(""))
            {
                searchStrings.Add("IS=" + searchParam_searchByISBN);
            }

            if (!searchParam_searchByTitle.Equals(""))
            {
                searchStrings.Add("TI=" + searchParam_searchByTitle);
            }

            if (!searchParam_searchByAuthor.Equals(""))
            {
                searchStrings.Add("AU=" + searchParam_searchByAuthor);
            }

            if (!searchParam_searchByPublisher.Equals(""))
            {
                searchStrings.Add("VL=" + searchParam_searchByPublisher);
            }

            if (!searchParam_searchBySubject.Equals(""))
            {
                searchStrings.Add("SW=" + searchParam_searchBySubject);
            }

            if (!searchParam_searchBySet.Equals(""))
            {
                searchStrings.Add("RH=" + searchParam_searchBySet);
            }

            if (!searchParam_searchByCollection.Equals(""))
            {
                searchStrings.Add("RH=" + searchParam_searchByCollection);
            }

            if (!searchParam_searchByProductForm.Equals(""))
            {
                searchStrings.Add("PF=" + searchParam_searchByProductForm);
            }

            if (!searchParam_searchByLanguage.Equals(""))
            {
                searchStrings.Add("SP=" + searchParam_searchByLanguage);
            }

            if(!searchParam_searchByPriceFrom.Equals("") && !searchParam_searchByPriceTo.Equals(""))
            {
                searchStrings.Add("PR=" + searchParam_searchByPriceFrom + "^" + searchParam_searchByPriceTo);
            }
            else if (!searchParam_searchByPriceFrom.Equals(""))
            {
                searchStrings.Add("PR=" + searchParam_searchByPriceFrom + "^9999.99");
            }
            else if (!searchParam_searchByPriceTo.Equals(""))
            {
                searchStrings.Add("PR=0.00^" + searchParam_searchByPriceTo);
            }

            if (!searchParam_searchByPublicationDateFrom.Equals("") && !searchParam_searchByPublicationDateTo.Equals(""))
            {
                searchStrings.Add("EJ=" + searchParam_searchByPublicationDateFrom + "^" + searchParam_searchByPublicationDateTo);
            }
            else if (!searchParam_searchByPublicationDateFrom.Equals(""))
            {
                searchStrings.Add("EJ=" + searchParam_searchByPublicationDateFrom + "^2150");
            }
            else if (!searchParam_searchByPublicationDateTo.Equals(""))
            {
                searchStrings.Add("EJ=1750^" + searchParam_searchByPublicationDateTo);
            }

            request = string.Join(" AND ", searchStrings.ToArray<string>());

            switch (comboBox_search_database.SelectedIndex)
            {
                case 0:
                    searchParam_database = DATABASES.VLB;
                    break;
                case 1:
                    searchParam_database = DATABASES.SBZ;
                    break;
                case 2:
                    searchParam_database = DATABASES.OESB;
                    break;
            }

            //Init new RestClient
            client = new RestClient();
            client.EndPoint = string.Format(@"http://api.vlb.de/api/v1/products?page={3}&size={4}&sort={0}&direction={1}&search={2}&source={5}", sortField, sortDirection, request, page, amountOfTitlesShown, searchParam_database.ToString());
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json-short";
            client.Method = HttpVerb.GET;

            //Send Request
            try
            {
                json = client.MakeRequest("", ref responseAsString);
            }
            catch (Exception ex)
            {
                toolStripStatusLabel.Text = String.Format("Fehler: {0}", ex.Message);
                return;
            }

            //Init Json.net Deserializer
            JsonSerializerSettings jsonSerializerSettings = new JsonSerializerSettings();
            jsonSerializerSettings.MissingMemberHandling = MissingMemberHandling.Ignore;
            jsonSerializerSettings.NullValueHandling = NullValueHandling.Ignore;

            //Debug Output
            if (!checkBox_debug_appendText.Checked)
            {
                richTextBox_debug_rawData.Clear();
            }

            richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
            richTextBox_debug_rawData.AppendText("Response:\n" + JsonFormatter.FormatJson(responseAsString) + "\n\n");

            //Create a Container, containing all Json Fields and deserialize Json Data
            JsonDataContainerSearch jsonDataContainerSearch = JsonConvert.DeserializeObject<JsonDataContainerSearch>(responseAsString, jsonSerializerSettings);

            //Init Data Grid View and Fill with Json Data
            dataGridView_search_results.Rows.Clear();

            if (jsonDataContainerSearch.content.Length == 0)
            {
                toolStripStatusLabel.Text = "Suche durchgeführt: 0 Treffer.";
                return;
            }

            dataGridView_search_results.Rows.Add(jsonDataContainerSearch.content.Length);

            for (int i = 0; i < jsonDataContainerSearch.content.Length; i++)
            {
                //ISBN
                dataGridView_search_results[0, i].Value = jsonDataContainerSearch.content[i].isbn;

                //Title
                dataGridView_search_results[1, i].Value = jsonDataContainerSearch.content[i].title;

                //Author
                dataGridView_search_results[2, i].Value = jsonDataContainerSearch.content[i].author;

                //Publisher
                dataGridView_search_results[3, i].Value = jsonDataContainerSearch.content[i].publisher;

                //Price EUR D
                double priceEURD;

                if (double.TryParse(jsonDataContainerSearch.content[i].priceEurD.Replace(".", ","), out priceEURD))
                {
                    dataGridView_search_results[4, i].Value = priceEURD.ToString("N2");
                }
                else
                {
                    dataGridView_search_results[4, i].Value = "";
                }

                //ProductType
                dataGridView_search_results[5, i].Value = jsonDataContainerSearch.content[i].productType;
            }

            //Set Search Result Data Grid View
            lastPage = int.Parse(jsonDataContainerSearch.totalPages);
            setPageButtonStatus();

            toolStripStatusLabel.Text = String.Format("Suche durchgeführt: {0} Treffer.", jsonDataContainerSearch.totalElements);
        }



        /// <summary>
        /// Gets the Titles Metadata.
        /// </summary>
        /// <param name="isbn">ISBN to search for.</param>
        /// <returns>The JsonDataContainer containing all available Metadata.</returns>
        private JsonDataContainerDetail getTitleDetail(string isbn)
        {
            //Variables
            RestClient client;
            HttpWebResponse json;
            string responseAsString = null;

            //Send REST Request
            client = new RestClient();
            client.EndPoint = @"http://api.vlb.de/api/v1/product/" + isbn.Replace("-", "") + @"/isbn13";
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.GET;

            //Send Request
            try
            {
                json = client.MakeRequest("", ref responseAsString);
            }
            catch (Exception ex)
            {
                throw ex;
            }

            //Init Json.net Deserializer
            JsonSerializerSettings jsonSerializerSettings = new JsonSerializerSettings();
            jsonSerializerSettings.MissingMemberHandling = MissingMemberHandling.Ignore;
            jsonSerializerSettings.NullValueHandling = NullValueHandling.Ignore;

            //Debug Output
            if (!checkBox_debug_appendText.Checked)
            {
                richTextBox_debug_rawData.Clear();
            }

            richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
            richTextBox_debug_rawData.AppendText("Response:\n" + JsonFormatter.FormatJson(responseAsString) + "\n\n");

            //Create a Container, containing all Json Fields and deserialize Json Data
            JsonDataContainerDetail jsonDataContainerDetailRaw = JsonConvert.DeserializeObject<JsonDataContainerDetail>(responseAsString, jsonSerializerSettings);
            return Transformer.transform(jsonDataContainerDetailRaw);
        }



        /// <summary>
        /// Gets all Publisher Data, if available.
        /// </summary>
        /// <param name="mvbid">The Publishers MVB ID.</param>
        private void getPublisherDetail(string mvbid)
        {
            //No MVB ID provided
            if (mvbid.Equals(""))
            {
                return;
            }

            //Variables
            RestClient client;
            HttpWebResponse json;
            List<string> searchStrings = new List<string>();
            string responseAsString = null;

            //Init new RestClient
            client = new RestClient();
            client.EndPoint = string.Format(@"http://api.vlb.de/api/v1/publisher/{0}", mvbid);
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.GET;

            //Send Request
            try
            {
                json = client.MakeRequest("", ref responseAsString);
            }
            catch (Exception ex)
            {
                richTextBox_debug_rawData.AppendText(String.Format("\r\rFehler: {0}\n\n", ex.Message));

                return;
            }

            //Debug
            richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
            richTextBox_debug_rawData.AppendText("Response:\n" + JsonFormatter.FormatJson(responseAsString) + "\n\n");

            //Init Json.net Deserializer
            JsonSerializerSettings jsonSerializerSettings = new JsonSerializerSettings();
            jsonSerializerSettings.MissingMemberHandling = MissingMemberHandling.Ignore;
            jsonSerializerSettings.NullValueHandling = NullValueHandling.Ignore;

            //Create a Container, containing all Json Fields and deserialize Json Data
            JsonDataContainerPublisher jsonDataContainerPublisher = JsonConvert.DeserializeObject<JsonDataContainerPublisher>(responseAsString, jsonSerializerSettings);

            //Set Publisher Detail View
            richTextBox_publisher_name.Text = jsonDataContainerPublisher.name;

            richTextBox_publisher_mvbid.Text = jsonDataContainerPublisher.mvbId;
            richTextBox_publisher_street.Text = jsonDataContainerPublisher.street;
            richTextBox_publisher_cityStreet.Text = jsonDataContainerPublisher.cityStreet;
            richTextBox_publisher_zipStreet.Text = jsonDataContainerPublisher.zipStreet;
            richTextBox_publisher_postbox.Text = jsonDataContainerPublisher.postbox;
            richTextBox_publisher_zipPostbox.Text = jsonDataContainerPublisher.zipPostbox;

            richTextBox_publisher_phone.Text = jsonDataContainerPublisher.phone;
            richTextBox_publisher_fax.Text = jsonDataContainerPublisher.fax;
            linkLabel_publisher_email.Text = jsonDataContainerPublisher.email;
            linkLabel_publisher_website.Text = jsonDataContainerPublisher.url;

            richTextBox_publisher_deliveryBS.Text = jsonDataContainerPublisher.deliveryBS;

            if(jsonDataContainerPublisher.isbnPrefixes.Length > 0)
            {
                linkLabel_publisher_isbnPrefixes1.Text = jsonDataContainerPublisher.isbnPrefixes[0];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 1)
            {
                linkLabel_publisher_isbnPrefixes2.Text = jsonDataContainerPublisher.isbnPrefixes[1];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 2)
            {
                linkLabel_publisher_isbnPrefixes3.Text = jsonDataContainerPublisher.isbnPrefixes[2];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 3)
            {
                linkLabel_publisher_isbnPrefixes4.Text = jsonDataContainerPublisher.isbnPrefixes[3];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 4)
            {
                linkLabel_publisher_isbnPrefixes5.Text = jsonDataContainerPublisher.isbnPrefixes[4];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 5)
            {
                linkLabel_publisher_isbnPrefixes6.Text = jsonDataContainerPublisher.isbnPrefixes[5];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 6)
            {
                linkLabel_publisher_isbnPrefixes7.Text = jsonDataContainerPublisher.isbnPrefixes[6];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 7)
            {
                linkLabel_publisher_isbnPrefixes8.Text = jsonDataContainerPublisher.isbnPrefixes[7];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 8)
            {
                linkLabel_publisher_isbnPrefixes9.Text = jsonDataContainerPublisher.isbnPrefixes[8];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 9)
            {
                linkLabel_publisher_isbnPrefixes10.Text = jsonDataContainerPublisher.isbnPrefixes[9];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 10)
            {
                linkLabel_publisher_isbnPrefixes11.Text = jsonDataContainerPublisher.isbnPrefixes[10];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 11)
            {
                linkLabel_publisher_isbnPrefixes12.Text = jsonDataContainerPublisher.isbnPrefixes[11];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 12)
            {
                linkLabel_publisher_isbnPrefixes13.Text = jsonDataContainerPublisher.isbnPrefixes[12];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 13)
            {
                linkLabel_publisher_isbnPrefixes14.Text = jsonDataContainerPublisher.isbnPrefixes[13];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 14)
            {
                linkLabel_publisher_isbnPrefixes15.Text = jsonDataContainerPublisher.isbnPrefixes[14];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 15)
            {
                linkLabel_publisher_isbnPrefixes16.Text = jsonDataContainerPublisher.isbnPrefixes[15];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 16)
            {
                linkLabel_publisher_isbnPrefixes17.Text = jsonDataContainerPublisher.isbnPrefixes[16];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 17)
            {
                linkLabel_publisher_isbnPrefixes18.Text = jsonDataContainerPublisher.isbnPrefixes[17];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 18)
            {
                linkLabel_publisher_isbnPrefixes19.Text = jsonDataContainerPublisher.isbnPrefixes[18];
            }

            if (jsonDataContainerPublisher.isbnPrefixes.Length > 19)
            {
                linkLabel_publisher_isbnPrefixes20.Text = jsonDataContainerPublisher.isbnPrefixes[19];
            }
        }



        /// <summary>
        /// Gets the Cover by ISBN, if available. NOT USED ANYMORE.
        /// </summary>
        /// <param name="isbn">ISBN to search for.</param>
        private void getCover(string isbn)
        {
            //Reset Cover
            if (pictureBox_detail_cover.Image != null)
            {
                pictureBox_detail_cover.Image.Dispose();
            }

            //Variables
            RestClient client;
            HttpWebResponse httpWebResponse;
            Stream stream;

            //Configure REST Request
            client = new RestClient();
            client.EndPoint = @"http://api.vlb.de/api/v1/cover/" + isbn.Replace("-", "") + @"/m";
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.GET;

            //Send Request and get Image, if possible
            try
            {
                httpWebResponse = client.MakeRequestReturnResponse("");
                stream = httpWebResponse.GetResponseStream();

                pictureBox_detail_cover.Image = Image.FromStream(stream);
            }
            catch (Exception ex)
            {
                //Cover unavailable
                pictureBox_detail_cover.Image = Image.FromFile(Application.StartupPath + "/bin/cover_vlb_missing.png");

                richTextBox_debug_rawData.AppendText(String.Format("\r\rFehler: {0}\n\n", ex.Message));
            }

            richTextBox_debug_rawData.AppendText("Request:\nEndpoint: " + client.EndPoint + "\nAccessToken: " + client.AccessToken + "\nContentType: " + client.ContentTypeString + "\nMethod: " + client.Method + "\n\n");
        }



        /// <summary>
        /// Gets the Cover of the currently shown Title in the TitleDetailView asynchronly.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void getCoverAsyncDoWork(object sender, EventArgs e)
        {
            //Variables
            RestClient client;
            HttpWebResponse httpWebResponse;
            Stream stream;

            //Configure REST Request
            client = new RestClient();
            client.EndPoint = @"http://api.vlb.de/api/v1/cover/" + currentlyShownISBN.Replace("-", "") + @"/m";
            client.AccessToken = string.Format("Bearer {0}", statefullAccessToken);
            client.ContentTypeString = "application/json";
            client.Method = HttpVerb.GET;

            //Send Request and get Image, if possible
            try
            {
                httpWebResponse = client.MakeRequestReturnResponse("");
                stream = httpWebResponse.GetResponseStream();

                cover = Image.FromStream(stream);
            }
            catch (Exception ex)
            {
                //Cover unavailable
                cover = Image.FromFile("bin/cover_vlb_missing.png");
            }

        }



        /// <summary>
        /// Called, when the async Cover Download has finished. Shows the Cover in the TitleDetailView.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void getCoverAsyncCompleted(object sender, EventArgs e)
        {
            pictureBox_detail_cover.Image = cover;
            pictureBox_detail_cover.Refresh();
        }



        /// <summary>
        /// Creates a new Label, which can be attached to any Form.
        /// </summary>
        /// <param name="location"></param>
        /// <param name="size"></param>
        /// <param name="name"></param>
        /// <param name="text"></param>
        /// <param name="bold"></param>
        /// <returns></returns>
        private Label getNewFormattedLabel(Point location, Size size, string name, string text, bool bold)
        {
            Label label = new Label();
            label.AutoEllipsis = true;
            label.Font = new Font("Microsoft Sans Serif", 8.25F, bold ? FontStyle.Bold : FontStyle.Regular, GraphicsUnit.Point, ((byte)(0)));
            label.Location = location;
            label.MaximumSize = size;
            label.Name = name;
            label.Size = size;
            label.TabIndex = 999;
            label.Text = text;

            return label;
        }



        /// <summary>
        /// Creates a new LinkLabel, which can be attached to any Form.
        /// </summary>
        /// <param name="location"></param>
        /// <param name="size"></param>
        /// <param name="name"></param>
        /// <param name="text"></param>
        /// <param name="bold"></param>
        /// <param name="booleanSearchField"></param>
        /// <returns></returns>
        private LinkLabel getNewFormattedLinkLabel(Point location, Size size, string name, string text, bool bold, string booleanSearchField)
        {
            LinkLabel label = new LinkLabel();
            label.AutoEllipsis = true;
            label.Font = new Font("Microsoft Sans Serif", 8.25F, bold ? FontStyle.Bold : FontStyle.Regular, GraphicsUnit.Point, ((byte)(0)));
            label.Location = location;
            label.MaximumSize = size;
            label.Name = name;
            label.Size = size;
            label.TabIndex = 999;
            label.Text = text;
            label.Links[0].Name = booleanSearchField;
            label.LinkClicked += new LinkLabelLinkClickedEventHandler(linkLabel_LinkClicked);

            return label;
        }



        /// <summary>
        /// Creates a new RichTextBox, which can be attached to any Form.
        /// </summary>
        /// <param name="location"></param>
        /// <param name="size"></param>
        /// <param name="name"></param>
        /// <param name="text"></param>
        /// <param name="bold"></param>
        /// <param name="variableHeight"></param>
        /// <returns></returns>
        private RichTextBox getNewFormattedRichTextBox(Point location, Size size, string name, string text, bool bold, bool variableHeight)
        {
            RichTextBox richTextBox = new RichTextBox();
            richTextBox.BackColor = Color.White;
            richTextBox.BorderStyle = BorderStyle.None;
            richTextBox.Font = new Font("Microsoft Sans Serif", 8.25F, bold ? FontStyle.Bold : FontStyle.Regular, GraphicsUnit.Point, ((byte)(0)));
            richTextBox.Location = location;
            richTextBox.Name = name;
            richTextBox.ReadOnly = true;
            richTextBox.TabIndex = 999;
            richTextBox.Text = text;
            richTextBox.WordWrap = true;
            richTextBox.DetectUrls = false;

            if (variableHeight)
            {
                Size textSize = TextRenderer.MeasureText(text, richTextBox.Font, size, TextFormatFlags.WordBreak);
                richTextBox.MaximumSize = textSize;
                richTextBox.Size = textSize;
            }
            else
            {
                richTextBox.MaximumSize = size;
                richTextBox.Size = size;
            }

            return richTextBox;
        }



        /// <summary>
        /// Arranges a set of Data in a pseudo-Table View.
        /// </summary>
        /// <param name="data">The Data to show, as List of Lists (X / Y Coordinates, Origin is upper-left Corner).</param>
        /// <param name="font">Fonts to use to show the Data.</param>
        /// <param name="hyperlinkValue">If a non-empty String is set, the Text will be displayed in as a LinkLabel. The provided Hyperlink will be passed to the Listener.</param>
        /// <param name="columnWidth">Width of each Column (Row Height is variable).</param>
        /// <param name="columnWordWrap">True, if this Columns Rows should be multilined, else false.</param>
        /// <param name="spacingX">Space (Width) between each Element.</param>
        /// <param name="spacingY">Space (Height) between each Element.</param>
        /// <param name="offsetX">Offset from the upper-left Corner.</param>
        /// <param name="offsetY">Offset from the upper-left Corner.</param>
        /// <returns></returns>
        private List<Control> arrangeDataInTable(string[,] data, Font[,] font, string[,] hyperlinkValue, int[] columnWidth, bool[] columnWordWrap, int spacingX, int spacingY, int offsetX, int offsetY)
        {
            //Variables
            List<Control> controls = new List<Control>();
            int rowCurrentMaximumHeight = 0;
            int rowHeightSum = 0;
            int columnWidthSum = 0;

            //Loop over Rows
            for (int y = 0; y < data.GetLength(1); y++)
            {
                //Loop over Columns
                for(int x = 0; x < data.GetLength(0); x++)
                {
                    if (hyperlinkValue[x, y].Equals(""))
                    {
                        //Generate RichTextBox
                        RichTextBox richTextBox = new RichTextBox();
                        richTextBox.BackColor = Color.WhiteSmoke;
                        richTextBox.BorderStyle = BorderStyle.None;
                        richTextBox.Font = font[x, y];
                        richTextBox.Location = new Point(offsetX + spacingX * (x + 1) + columnWidthSum, offsetY + spacingY * (y + 1) + rowHeightSum);
                        richTextBox.Name = "richTextBox_" + data[x, y];
                        richTextBox.ReadOnly = true;
                        richTextBox.TabIndex = 999;
                        richTextBox.WordWrap = columnWordWrap[x];
                        richTextBox.ScrollBars = RichTextBoxScrollBars.None;
                        richTextBox.DetectUrls = false;

                        //Calculate Control Height
                        if (columnWordWrap[x])
                        {
                            Size textSize = TextRenderer.MeasureText(data[x, y], richTextBox.Font, new Size(columnWidth[x], 1), TextFormatFlags.WordBreak);
                            richTextBox.Text = data[x, y];
                            richTextBox.MaximumSize = new Size(columnWidth[x], textSize.Height);
                            richTextBox.Size = new Size(columnWidth[x], textSize.Height);
                        }
                        else
                        {
                            Size textSize = TextRenderer.MeasureText(data[x, y], richTextBox.Font, new Size(columnWidth[x], 1), TextFormatFlags.EndEllipsis);
                            richTextBox.Text = truncateTextWithEllipsis(data[x, y], font[x, y], columnWidth[x]);
                            richTextBox.MaximumSize = new Size(columnWidth[x], textSize.Height);
                            richTextBox.Size = new Size(columnWidth[x], textSize.Height);
                        }

                        controls.Add(richTextBox);

                        //Caulculate Column Width
                        columnWidthSum += columnWidth[x];

                        //Calculate Row Height 1/2
                        if (rowCurrentMaximumHeight < richTextBox.Size.Height)
                        {
                            rowCurrentMaximumHeight = richTextBox.Size.Height;
                        }

                    }
                    else
                    {
                        //Generate LinkLabel
                        LinkLabel label = new LinkLabel();
                        label.AutoEllipsis = false;
                        label.BackColor = Color.WhiteSmoke;
                        label.BorderStyle = BorderStyle.None;
                        label.Font = font[x, y];
                        label.Location = new Point(offsetX + spacingX * (x + 1) + columnWidthSum, offsetY + spacingY * (y + 1) + rowHeightSum);
                        label.Name = "linkLabel_" + data[x, y];
                        label.TabIndex = 999;
                        label.Links[0].Name = hyperlinkValue[x, y];
                        label.LinkClicked += new LinkLabelLinkClickedEventHandler(linkLabel_LinkClicked);

                        //Calculate Control Height
                        if (columnWordWrap[x])
                        {
                            Size textSize = TextRenderer.MeasureText(data[x, y], label.Font, new Size(columnWidth[x], 1), TextFormatFlags.WordBreak);
                            label.Text = data[x, y];
                            label.MaximumSize = new Size(columnWidth[x], textSize.Height);
                            label.Size = new Size(columnWidth[x], textSize.Height);
                        }
                        else
                        {
                            Size textSize = TextRenderer.MeasureText(data[x, y], label.Font, new Size(columnWidth[x], 1), TextFormatFlags.EndEllipsis);
                            label.Text = truncateTextWithEllipsis(data[x, y], font[x, y], columnWidth[x]);
                            label.MaximumSize = new Size(columnWidth[x], textSize.Height);
                            label.Size = new Size(columnWidth[x], textSize.Height);
                        }

                        controls.Add(label);

                        //Caulculate Column Width
                        columnWidthSum += columnWidth[x];

                        //Calculate Row Height 1/2
                        if (rowCurrentMaximumHeight < label.Size.Height)
                        {
                            rowCurrentMaximumHeight = label.Size.Height;
                        }

                    }
                    
                }

                columnWidthSum = 0;

                //Calculate Row Height 2/2
                rowHeightSum += rowCurrentMaximumHeight;
                rowCurrentMaximumHeight = 0;
            }

            return controls;
        }



        /// <summary>
        /// Truncates the provided Text, so it fits in to a Form with the specified Width.
        /// </summary>
        /// <param name="text">Text to fit into the Form.</param>
        /// <param name="font">Font to use.</param>
        /// <param name="formWidth">Width of the Form.</param>
        /// <returns>The truncated Text with Ellipsis Characters (...). Returns an empty String, if the Text did not fit at all. Returns null, if a null Input Text was provided.</returns>
        private string truncateTextWithEllipsis(string text, Font font, int formWidth)
        {
            //Input was null
            if(text == null)
            {
                return null;
            }

            //Variables
            int calculatedMaximumLength = 0;
            int segmentLength = text.Length;
            int nextSegmentLength = 0;

            //Initial Check: does the Text need any Truncate at all?
            if(TextRenderer.MeasureText(text, font, new Size(formWidth, 1), TextFormatFlags.SingleLine).Width <= formWidth)
            {
                return text;
            }

            //Keep running until perfectly fitting Size has been calculated
            while(segmentLength > 1)
            {
                //Calculate Segment Length
                segmentLength -= segmentLength / 2;
                nextSegmentLength = calculatedMaximumLength + segmentLength;

                //Check, if Text fits within formWidth
                if(TextRenderer.MeasureText(text.Substring(0, nextSegmentLength) + "...", font, new Size(formWidth, 1), TextFormatFlags.SingleLine).Width <= formWidth)
                {
                    calculatedMaximumLength += segmentLength;
                }
            }

            //Text did not fit at all - will probably never happen
            if(calculatedMaximumLength == 0)
            {
                return "";
            }

            return text.Substring(0, calculatedMaximumLength) + "...";
        }



        /// <summary>
        /// Exports a specific Collection of Metadata for given ISBNs as .txt Textfile.
        /// </summary>
        /// <param name="ISBNs">A List of ISBNs to export.</param>
        private void export_asTXT(List<string> ISBNs)
        {
            //Variables
            List<JsonDataContainerDetail> jsonDataContainerDetailList = new List<JsonDataContainerDetail>();
            string isbn13 = "";
            string title = "";
            string subtitle = "";
            string authors = "";
            string publisher = "";
            string extent = "";
            string weight = "";
            string pages = "";
            string productType = "";
            string binding = "";
            string prices = "";

            //Save exported File (show saveFileDialog)
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "txt|*.txt";
            saveFileDialog.InitialDirectory = Application.StartupPath;

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                //Get Data from VLB
                try
                {
                    for (int i = 0; i < ISBNs.Count; i++)
                    {
                        jsonDataContainerDetailList.Add(getTitleDetail(ISBNs[i]));
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                    return;
                }

                //Assemble Metadata for Export
                for (int i = 0; i < ISBNs.Count; i++)
                {
                    #region ISBN13
                    for (int j = 0; j < jsonDataContainerDetailList[i].identifiers.Length; j++)
                    {
                        if (jsonDataContainerDetailList[i].identifiers[j].type.Equals("15", StringComparison.CurrentCultureIgnoreCase))
                        {
                            isbn13 = jsonDataContainerDetailList[i].identifiers[j].value;
                            break;
                        }
                    }
                    #endregion

                    #region Title
                    title = jsonDataContainerDetailList[i].titles[0].title;
                    #endregion

                    #region Subtitle
                    subtitle = jsonDataContainerDetailList[i].titles[0].subtitle;
                    #endregion

                    #region Authors
                    List<string> authorsList = new List<string>();

                    //Get Contributor Names
                    for (int j = 0; j < jsonDataContainerDetailList[i].contributors.Length; j++)
                    {
                        authorsList.Add(String.Format("{0} {1}", jsonDataContainerDetailList[i].contributors[j].firstName, jsonDataContainerDetailList[i].contributors[j].lastName));
                    }

                    authors = string.Join(", ", authorsList.ToArray<string>());
                    #endregion

                    #region Publisher
                    for (int j = 0; j < jsonDataContainerDetailList[i].publishers.Length; j++)
                    {
                        if (jsonDataContainerDetailList[i].publishers[j].type.Equals("01", StringComparison.CurrentCultureIgnoreCase))
                        {
                            publisher = jsonDataContainerDetailList[i].publishers[j].name;
                        }
                    }
                    #endregion

                    #region Pages
                    if (!jsonDataContainerDetailList[i].extent.pages.Equals(""))
                    {
                        pages = string.Format("{0} S.", jsonDataContainerDetailList[i].extent.pages);
                    }
                    #endregion

                    #region Extent
                    if (!jsonDataContainerDetailList[i].form.height.Equals("") && !jsonDataContainerDetailList[i].form.width.Equals(""))
                    {
                        extent = string.Format("{0} x {1} mm", jsonDataContainerDetailList[i].form.height, jsonDataContainerDetailList[i].form.width);
                    }
                    #endregion

                    #region Weight
                    if (!jsonDataContainerDetailList[i].form.weight.Equals(""))
                    {
                        weight = string.Format("{0} g", jsonDataContainerDetailList[i].form.weight);
                    }
                    #endregion

                    #region ProductType
                    if (!jsonDataContainerDetailList[i].form.type.Equals(""))
                    {
                        productType = codeList_007_productform.getCodeListValue(jsonDataContainerDetailList[i].form.type);
                    }
                    #endregion

                    #region Binding
                    if (!jsonDataContainerDetailList[i].form.binding.Equals(""))
                    {
                        binding = jsonDataContainerDetailList[i].form.binding;
                    }
                    #endregion

                    #region Price
                    List<string> pricesList = new List<string>();

                    for (int j = 0; j < jsonDataContainerDetailList[i].prices.Length; j++)
                    {
                        pricesList.Add(string.Format(@"{0} {1} [{2} / {3}]", String.Format("{0}{1}", jsonDataContainerDetailList[i].prices[j].currency, jsonDataContainerDetailList[i].prices[j].country.Equals("") ? "" : string.Format(" ({0})", jsonDataContainerDetailList[i].prices[j].country)), jsonDataContainerDetailList[i].prices[j].value, jsonDataContainerDetailList[i].prices[j].taxRate, jsonDataContainerDetailList[i].prices[j].fixedRetailPrice));

                        if (!jsonDataContainerDetailList[i].prices[j].validFrom.Equals(""))
                        {
                            pricesList[pricesList.Count - 1] = string.Format("{0} (gültig ab: {1})", pricesList[pricesList.Count - 1], jsonDataContainerDetailList[i].prices[j].validFrom);
                        }

                        if (!jsonDataContainerDetailList[i].prices[j].validUntil.Equals(""))
                        {
                            pricesList[pricesList.Count - 1] = string.Format("{0} (gültig bis: {1})", pricesList[pricesList.Count - 1], jsonDataContainerDetailList[i].prices[j].validUntil);
                        }
                    }

                    prices = string.Join(", ", pricesList.ToArray<string>());
                    #endregion
                }

                //Save Data to File
                try
                {
                    //Create File
                    FileStream fileStream = new FileStream(saveFileDialog.FileName, FileMode.Create);
                    StreamWriter streamWriter = new StreamWriter(fileStream);

                    //Fill File with Metadata
                    for (int i = 0; i < ISBNs.Count; i++)
                    {
                        streamWriter.WriteLine("");
                        streamWriter.WriteLine(string.Format("       ISBN-13 : {0}", isbn13));
                        streamWriter.WriteLine(string.Format("         Titel : {0}", title));
                        streamWriter.WriteLine(string.Format("    Untertitel : {0}", subtitle));
                        streamWriter.WriteLine(string.Format("     Autor(en) : {0}", authors));
                        streamWriter.WriteLine(string.Format("        Verlag : {0}", publisher));
                        streamWriter.WriteLine(string.Format("        Seiten : {0}", pages));
                        streamWriter.WriteLine(string.Format("          Maße : {0}", extent));
                        streamWriter.WriteLine(string.Format("       Gewicht : {0}", weight));
                        streamWriter.WriteLine(string.Format("    Produkttyp : {0}", productType));
                        streamWriter.WriteLine(string.Format("    Einbandart : {0}", binding));
                        streamWriter.WriteLine(string.Format("         Preis : {0}", prices));
                        streamWriter.WriteLine("");
                    }

                    //Close File
                    streamWriter.Flush();
                    streamWriter.Close();
                    fileStream.Close();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                    return;
                }
            }
        }



        /// <summary>
        /// Exports a List of ISBNs as .txt Textfile, containing Orders to send to a Deliverer.
        /// </summary>
        /// <param name="ISBNs">A List of ISBNs to export.</param>
        private void export_order(List<string> ISBNs)
        {
            //Save exported File (show saveFileDialog)
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "txt|*.txt";
            saveFileDialog.InitialDirectory = Application.StartupPath;

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    //Create File
                    FileStream fileStream = new FileStream(saveFileDialog.FileName, FileMode.Create);
                    StreamWriter streamWriter = new StreamWriter(fileStream);

                    for(int i = 0; i < ISBNs.Count; i++)
                    {
                        streamWriter.WriteLine(string.Format("B101          VD30800     VD{0}383{1} {2}EN0001ST*9999", DateTime.Now.ToString("yyyyMMdd"), DateTime.Now.ToString("yyMMdd"), ISBNs[i]));
                    }

                    //Close File
                    streamWriter.Flush();
                    streamWriter.Close();
                    fileStream.Close();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                    return;
                }
            }
        }



        /// <summary>
        /// Creates a buchhandel.de QR-Code for the given ISBN and saves the resulting Bitmap at a given Location.
        /// 
        /// QR-Code Generator bases upon:
        /// http://code-bude.net/2013/10/17/qrcoder-eine-open-source-qr-code-implementierung-in-csharp/
        /// </summary>
        /// <param name="ISBN">The ISBN to use to create the QR-Code.</param>
        /// <param name="pathAndFileName">Path and Filename (inclusing FileNameExtension) to save the Bitmap. Use .png as FileNameExtension!</param>
        /// <returns>True, if the Export was successfull, false otherwise.</returns>
        private bool export_asQRCode(string ISBN, string pathAndFileName)
        {
            try
            {
                //Generate QR-Code
                QRCodeGenerator qrGenerator = new QRCodeGenerator();
                QRCodeData qrCodeData = qrGenerator.CreateQrCode(String.Format("https://www.buchhandel.de/buch/{0}?MVB-Kennnummer=5006132", ISBN), QRCodeGenerator.ECCLevel.H);
                QRCode qrCode = new QRCode(qrCodeData);
                Bitmap bitmap = qrCode.GetGraphic(20, Color.Black, Color.White, (Bitmap)Bitmap.FromFile(Application.StartupPath + @"\bin\qrcode_buchhandel_image.png"), 90, 3);

                //Save File
                bitmap.Save(pathAndFileName);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Ein Fehler ist aufgetreten:\n\n" + ex.Message, "Fehler");
                return false;
            }

            return true;
        }

        
    }
}
