﻿using System;
using System.IO;
using System.Net;
using System.Text;
using System.Windows.Forms;



public enum HttpVerb
{
    GET,
    POST,
    PUT,
    DELETE
}

namespace MVB_RestBookService
{

    /// <summary>
    /// Quelle:
    /// http://www.codeproject.com/Tips/497123/How-to-make-REST-requests-with-Csharp
    /// </summary>
    public class RestClient
    {
        public string EndPoint { get; set; }
        public HttpVerb Method { get; set; }
        public string ContentType { get; set; }
        public string PostData { get; set; }
        public string AccessToken { get; set; }
        public string ContentTypeString { get; set; }



        public RestClient()
        {
            EndPoint = "";
            Method = HttpVerb.GET;
            ContentType = "text/xml";
            PostData = "";
        }

        public RestClient(string endpoint)
        {
            EndPoint = endpoint;
            Method = HttpVerb.GET;
            ContentType = "text/xml";
            PostData = "";
        }

        public RestClient(string endpoint, HttpVerb method)
        {
            EndPoint = endpoint;
            Method = method;
            ContentType = "text/xml";
            PostData = "";
        }

        public RestClient(string endpoint, HttpVerb method, string postData)
        {
            EndPoint = endpoint;
            Method = method;
            ContentType = "text/xml";
            PostData = postData;
        }



        public HttpWebResponse MakeRequest(string parameters, ref string responseAsString)
        {
            var request = (HttpWebRequest)WebRequest.Create(EndPoint + parameters);

            request.Method = Method.ToString();
            request.Headers.Set(HttpRequestHeader.Authorization, AccessToken);
            request.ContentType = ContentTypeString;
            request.Accept = ContentTypeString;

            if (!string.IsNullOrEmpty(PostData) && Method == HttpVerb.POST)
            {
                var encoding = new UTF8Encoding();
                var bytes = Encoding.GetEncoding("iso-8859-1").GetBytes(PostData);
                request.ContentLength = bytes.Length;

                using (var writeStream = request.GetRequestStream())
                {
                    writeStream.Write(bytes, 0, bytes.Length);
                }
            }

            using (var response = (HttpWebResponse)request.GetResponse())
            {
                var responseValue = string.Empty;
                
                if (response.StatusCode != HttpStatusCode.OK)
                {
                    var message = String.Format("Request failed. Received HTTP {0}", response.StatusCode);
                    throw new ApplicationException(message);
                }

                // grab the response
                using (var responseStream = response.GetResponseStream())
                {
                    if (responseStream != null)
                        using (var reader = new StreamReader(responseStream))
                        {
                            responseValue = reader.ReadToEnd();
                        }
                }

                responseAsString = responseValue;

                for(int i = 0; i < response.Headers.Keys.Count; i++)
                {
                    //MessageBox.Show(response.Headers.Keys[i] + "   " + response.Headers[response.Headers.Keys[i]]);
                }

                return response;
            }
        }

        public HttpWebResponse MakeRequestReturnResponse(string parameters)
        {

            var request = (HttpWebRequest)WebRequest.Create(EndPoint + parameters);

            request.Method = Method.ToString();
            request.Headers.Set(HttpRequestHeader.Authorization, AccessToken);
            request.ContentType = ContentTypeString;
            request.Accept = ContentTypeString;

            if (!string.IsNullOrEmpty(PostData) && Method == HttpVerb.POST)
            {
                var encoding = new UTF8Encoding();
                var bytes = Encoding.GetEncoding("iso-8859-1").GetBytes(PostData);
                request.ContentLength = bytes.Length;

                using (var writeStream = request.GetRequestStream())
                {
                    writeStream.Write(bytes, 0, bytes.Length);
                }
            }
            
            var response = (HttpWebResponse)request.GetResponse();

            if (response.StatusCode != HttpStatusCode.OK)
            {
                var message = String.Format("Request failed. Received HTTP {0}", response.StatusCode);
                throw new ApplicationException(message);
            }

            return response;
        }

    } // class

}